@@ -1,17 +1,5 @@
-#  Copyright 2023 MTS (Mobile Telesystems)
-#
-#  Licensed under the Apache License, Version 2.0 (the "License");
-#  you may not use this file except in compliance with the License.
-#  You may obtain a copy of the License at
-#
-#      http://www.apache.org/licenses/LICENSE-2.0
-#
-#  Unless required by applicable law or agreed to in writing, software
-#  distributed under the License is distributed on an "AS IS" BASIS,
-#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-#  See the License for the specific language governing permissions and
-#  limitations under the License.
-
+# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
+# SPDX-License-Identifier: Apache-2.0
 from onetl.base.base_connection import BaseConnection
 from onetl.base.base_db_connection import BaseDBConnection, BaseDBDialect
 from onetl.base.base_file_connection import BaseFileConnection
@@ -25,7 +13,7 @@ from onetl.base.base_file_format import BaseReadableFileFormat, BaseWritableFile
 from onetl.base.base_file_limit import BaseFileLimit
 from onetl.base.contains_exception import ContainsException
 from onetl.base.contains_get_df_schema import ContainsGetDFSchemaMethod
-from onetl.base.contains_get_min_max_bounds import ContainsGetMinMaxBounds
+from onetl.base.contains_get_min_max_values import ContainsGetMinMaxValues
 from onetl.base.path_protocol import PathProtocol, PathWithStatsProtocol
 from onetl.base.path_stat_protocol import PathStatProtocol
 from onetl.base.pure_path_protocol import PurePathProtocol
