@@ -1,17 +1,5 @@
-#  Copyright 2023 MTS (Mobile Telesystems)
-#
-#  Licensed under the Apache License, Version 2.0 (the "License");
-#  you may not use this file except in compliance with the License.
-#  You may obtain a copy of the License at
-#
-#      http://www.apache.org/licenses/LICENSE-2.0
-#
-#  Unless required by applicable law or agreed to in writing, software
-#  distributed under the License is distributed on an "AS IS" BASIS,
-#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-#  See the License for the specific language governing permissions and
-#  limitations under the License.
-
+# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
+# SPDX-License-Identifier: Apache-2.0
 from __future__ import annotations
 
 import os
@@ -27,13 +15,17 @@ from onetl.base.path_stat_protocol import PathStatProtocol
 
 class BaseFileConnection(BaseConnection):
     """
-    Implements generic methods for files and directories manipulation on some filesystem (usually remote)
+    Implements generic methods for files and directories manipulation on some filesystem (usually remote).
+
+    .. versionadded:: 0.8.0
     """
 
     @abstractmethod
     def path_exists(self, path: os.PathLike | str) -> bool:
         """
-        Check if specified path exists on remote filesystem. |support_hooks|
+        Check if specified path exists on remote filesystem. |support_hooks|.
+
+        .. versionadded:: 0.8.0
 
         Parameters
         ----------
@@ -47,11 +39,12 @@ class BaseFileConnection(BaseConnection):
         Examples
         --------
 
-        .. code:: python
-
-            assert connection.path_exists("/path/to/file.csv")
-            assert connection.path_exists("/path/to/dir")
-            assert not connection.path_exists("/path/to/missing")
+        >>> connection.path_exists("/path/to/file.csv")
+        True
+        >>> connection.path_exists("/path/to/dir")
+        True
+        >>> connection.path_exists("/path/to/missing")
+        False
         """
 
     @abstractmethod
@@ -59,6 +52,8 @@ class BaseFileConnection(BaseConnection):
         """
         Check if specified path is a file. |support_hooks|
 
+        .. versionadded:: 0.8.0
+
         Parameters
         ----------
         path : str or :obj:`os.PathLike`
@@ -76,10 +71,10 @@ class BaseFileConnection(BaseConnection):
         Examples
         --------
 
-        .. code:: python
-
-            assert connection.is_file("/path/to/dir/file.csv")
-            assert not connection.is_file("/path/to/dir")
+        >>> connection.is_file("/path/to/dir/file.csv")
+        True
+        >>> connection.is_file("/path/to/dir")
+        False
         """
 
     @abstractmethod
@@ -87,6 +82,8 @@ class BaseFileConnection(BaseConnection):
         """
         Check if specified path is a directory. |support_hooks|
 
+        .. versionadded:: 0.8.0
+
         Parameters
         ----------
         path : str or :obj:`os.PathLike`
@@ -104,10 +101,10 @@ class BaseFileConnection(BaseConnection):
         Examples
         --------
 
-        .. code:: python
-
-            assert connection.is_dir("/path/to/dir")
-            assert not connection.is_dir("/path/to/dir/file.csv")
+        >>> connection.is_dir("/path/to/dir")
+        True
+        >>> connection.is_dir("/path/to/dir/file.csv")
+        False
         """
 
     @abstractmethod
@@ -115,6 +112,8 @@ class BaseFileConnection(BaseConnection):
         """
         Returns stats for a specific path. |support_hooks|
 
+        .. versionadded:: 0.8.0
+
         Parameters
         ----------
         path : str or :obj:`os.PathLike`
@@ -131,11 +130,11 @@ class BaseFileConnection(BaseConnection):
         Examples
         --------
 
-        .. code:: python
-
-            stat = connection.get_stat("/path/to/file.csv")
-            assert stat.st_size > 0
-            assert stat.st_uid == 12345  # owner id
+        >>> stat = connection.get_stat("/path/to/file.csv")
+        >>> stat.st_size  # in bytes
+        1024
+        >>> stat.st_uid  # owner id or name
+        12345
         """
 
     @abstractmethod
@@ -143,6 +142,8 @@ class BaseFileConnection(BaseConnection):
         """
         Returns directory at specific path, with stats. |support_hooks|
 
+        .. versionadded:: 0.8.0
+
         Parameters
         ----------
         path : str or :obj:`os.PathLike`
@@ -163,11 +164,11 @@ class BaseFileConnection(BaseConnection):
         Examples
         --------
 
-        .. code:: python
-
-            dir_path = connection.resolve_dir("/path/to/dir")
-            assert os.fspath(dir_path) == "/path/to/dir"
-            assert dir_path.stat.st_uid == 12345  # owner id
+        >>> dir_path = connection.resolve_dir("/path/to/dir")
+        >>> os.fspath(dir_path)
+        '/path/to/dir'
+        >>> dir_path.stat().st_uid  # owner id
+        12345
         """
 
     @abstractmethod
@@ -175,6 +176,8 @@ class BaseFileConnection(BaseConnection):
         """
         Returns file at specific path, with stats. |support_hooks|
 
+        .. versionadded:: 0.8.0
+
         Parameters
         ----------
         path : str or :obj:`os.PathLike`
@@ -195,11 +198,11 @@ class BaseFileConnection(BaseConnection):
         Examples
         --------
 
-        .. code:: python
-
-            file_path = connection.resolve_file("/path/to/dir/file.csv")
-            assert os.fspath(file_path) == "/path/to/dir/file.csv"
-            assert file_path.stat.st_uid == 12345  # owner id
+        >>> file_path = connection.resolve_file("/path/to/dir/file.csv")
+        >>> os.fspath(file_path)
+        '/path/to/dir/file.csv'
+        >>> file_path.stat().st_uid  # owner id
+        12345
         """
 
     @abstractmethod
@@ -207,6 +210,8 @@ class BaseFileConnection(BaseConnection):
         """
         Creates directory tree on remote filesystem. |support_hooks|
 
+        .. versionadded:: 0.8.0
+
         Parameters
         ----------
         path : str or :obj:`os.PathLike`
@@ -224,9 +229,9 @@ class BaseFileConnection(BaseConnection):
         Examples
         --------
 
-        .. code:: python
-
-            dir_path = connection.create_dir("/path/to/dir")
+        >>> dir_path = connection.create_dir("/path/to/dir")
+        >>> os.fspath(dir_path)
+        '/path/to/dir'
         """
 
     @abstractmethod
@@ -240,6 +245,8 @@ class BaseFileConnection(BaseConnection):
 
             Supports only one file removal per call. Directory removal is **NOT** supported, use :obj:`~remove_dir` instead.
 
+        .. versionadded:: 0.8.0
+
         Parameters
         ----------
         path : str or :obj:`os.PathLike`
@@ -257,12 +264,12 @@ class BaseFileConnection(BaseConnection):
         Examples
         --------
 
-        .. code:: python
-
-            assert connection.remove_file("/path/to/file.csv")
-            assert not connection.path_exists("/path/to/dir/file.csv")
-
-            assert not connection.remove_file("/path/to/file.csv")  # already deleted
+        >>> connection.remove_file("/path/to/file.csv")
+        True
+        >>> connection.path_exists("/path/to/dir/file.csv")
+        False
+        >>> connection.remove_file("/path/to/file.csv")  # already deleted, no error
+        False
         """
 
     @abstractmethod
@@ -272,13 +279,17 @@ class BaseFileConnection(BaseConnection):
 
         If directory does not exist, no exception is raised.
 
+        .. versionadded:: 0.8.0
+
         Parameters
         ----------
         path : str or :obj:`os.PathLike`
-            Directory path to remote
+            Directory path to remove
 
         recursive : bool, default ``False``
-            If ``True``, remove directory tree recursively.
+            If ``True``, remove directory tree recursively (including files and subdirectories).
+
+            If ``False``, remove only directory itself. Directory should be empty.
 
         Returns
         -------
@@ -289,16 +300,24 @@ class BaseFileConnection(BaseConnection):
         NotADirectoryError
             Path is not a directory
 
+        :obj:`onetl.exception.DirectoryNotEmptyError`
+            Directory is not empty, and ``recursive`` is ``False``
+
         Examples
         --------
 
-        .. code:: python
-
-            assert connection.remove_dir("/path/to/dir")
-            assert not connection.path_exists("/path/to/dir/file.csv")
-            assert not connection.path_exists("/path/to/dir")
-
-            assert not connection.remove_dir("/path/to/dir")  # already deleted
+        >>> connection.remove_dir("/path/to/dir")
+        Traceback (most recent call last):
+            ...
+        onetl.exception.DirectoryNotEmptyError: Cannot delete non-empty directory '/path/to/dir'
+        >>> connection.remove_dir("/path/to/dir", recirsive=True)
+        True
+        >>> connection.path_exists("/path/to/dir")
+        False
+        >>> connection.path_exists("/path/to/dir/file.csv")
+        False
+        >>> connection.remove_dir("/path/to/dir")  # already deleted, no error
+        False
         """
 
     @abstractmethod
@@ -315,6 +334,8 @@ class BaseFileConnection(BaseConnection):
 
             Supports only one file move per call. Directory move/rename is **NOT** supported.
 
+        .. versionadded:: 0.8.0
+
         Parameters
         ----------
         source_file_path : str or :obj:`os.PathLike`
@@ -344,11 +365,13 @@ class BaseFileConnection(BaseConnection):
         Examples
         --------
 
-        .. code:: python
-
-            new_file = connection.rename_file("/path/to/file1.csv", "/path/to/file2.csv")
-            assert connection.path_exists("/path/to/file2.csv")
-            assert not connection.path_exists("/path/to/file1.csv")
+        >>> new_file = connection.rename_file("/path/to/file1.csv", "/path/to/file2.csv")
+        >>> os.fspath(new_file)
+        '/path/to/file2.csv'
+        >>> connection.path_exists("/path/to/file2.csv")
+        True
+        >>> connection.path_exists("/path/to/file1.csv")
+        False
         """
 
     @abstractmethod
@@ -361,6 +384,11 @@ class BaseFileConnection(BaseConnection):
         """
         Return list of child files/directories in a specific directory. |support_hooks|
 
+        .. versionadded:: 0.8.0
+
+        .. versionchanged:: 0.14.0
+            Method returns full file/directory path instead of just name.
+
         Parameters
         ----------
         path : str or :obj:`os.PathLike`
@@ -388,11 +416,11 @@ class BaseFileConnection(BaseConnection):
         Examples
         --------
 
-        .. code:: python
-
-            dir_content = connection.list_dir("/path/to/dir")
-            assert os.fspath(dir_content[0]) == "/path/to/dir/file.csv"
-            assert connection.path_exists("/path/to/dir/file.csv")
+        >>> dir_content = connection.list_dir("/path/to/dir")
+        >>> os.fspath(dir_content[0])
+        '/path/to/dir/file.csv'
+        >>> connection.path_exists("/path/to/dir/file.csv")
+        True
         """
 
     @abstractmethod
@@ -408,6 +436,11 @@ class BaseFileConnection(BaseConnection):
 
         Just like :obj:`os.walk`, but with additional filter/limit logic.
 
+        .. versionadded:: 0.8.0
+
+        .. versionchanged:: 0.14.0
+            Method returns full file/directory path instead of just name.
+
         Parameters
         ----------
         root : str or :obj:`os.PathLike`
@@ -417,11 +450,11 @@ class BaseFileConnection(BaseConnection):
             If ``True``, walk in top-down order, otherwise walk in bottom-up order.
 
         filters : list of :obj:`BaseFileFilter <onetl.base.base_file_filter.BaseFileFilter>`, optional
-            Return only files/directories matching these filters. See :ref:`file-filters`
+            Return only files/directories matching these filters. See :ref:`file-filters`.
 
         limits : list of :obj:`BaseFileLimit <onetl.base.base_file_limit.BaseFileLimit>`, optional
-            Apply limits to the list of files/directories, and stop if one of the limits is reached.
-            See :ref:`file-limits`
+            Apply limits to the list of files/directories, and immediately stop if any of these limits is reached.
+            See :ref:`file-limits`.
 
         Returns
         -------
@@ -440,13 +473,16 @@ class BaseFileConnection(BaseConnection):
         Examples
         --------
 
-        .. code:: python
-
-            for root, dirs, files in connection.walk("/path/to/dir"):
-                assert os.fspath(root) == "/path/to/dir"
-                assert dirs == []
-                assert os.fspath(files[0]) == "/path/to/dir/file.csv"
-                assert connection.path_exists("/path/to/dir/file.csv")
+        >>> for root, dirs, files in connection.walk("/path/to/dir"):
+        ...    break
+        >>> os.fspath(root)
+        '/path/to/dir'
+        >>> dirs
+        []
+        >>> os.fspath(files[0])
+        '/path/to/dir/file.csv'
+        >>> connection.path_exists("/path/to/dir/file.csv")
+        True
         """
 
     @abstractmethod
@@ -463,6 +499,8 @@ class BaseFileConnection(BaseConnection):
 
             Supports only one file download per call. Directory download is **NOT** supported, use :ref:`file-downloader` instead.
 
+        .. versionadded:: 0.8.0
+
         Parameters
         ----------
         remote_file_path : str or :obj:`os.PathLike`
@@ -495,14 +533,18 @@ class BaseFileConnection(BaseConnection):
         Examples
         --------
 
-        .. code:: python
-
-            local_file = connection.download_file(
-                remote_file_path="/path/to/source.csv", local_file_path="/path/to/target.csv"
-            )
-            assert local_file.exists()
-            assert os.fspath(local_file) == "/path/to/target.csv"
-            assert local_file.stat().st_size == connection.get_stat("/path/to/source.csv").st_size
+        >>> local_file = connection.download_file(
+        ...     remote_file_path="/path/to/source.csv",
+        ...     local_file_path="/path/to/target.csv",
+        ... )
+        >>> os.fspath(local_file)
+        '/path/to/target.csv'
+        >>> local_file.exists()
+        True
+        >>> local_file.stat().st_size  # in bytes
+        1024
+        >>> connection.get_stat("/path/to/source.csv").st_size  # same size
+        1024
         """
 
     @abstractmethod
@@ -519,6 +561,8 @@ class BaseFileConnection(BaseConnection):
 
             Supports only one file upload per call. Directory upload is **NOT** supported, use :ref:`file-uploader` instead.
 
+        .. versionadded:: 0.8.0
+
         Parameters
         ----------
         local_file_path : str or :obj:`os.PathLike`
@@ -551,14 +595,18 @@ class BaseFileConnection(BaseConnection):
         Examples
         --------
 
-        .. code:: python
-
-            remote_file = connection.upload(
-                local_file_path="/path/to/source.csv",
-                remote_file_path="/path/to/target.csv",
-            )
-            assert connection.path_exists("/path/to/target.csv")
-            assert remote_file.stat().st_size == os.stat("/path/to/source.csv").st_size
+        >>> remote_file = connection.upload(
+        ...     local_file_path="/path/to/source.csv",
+        ...     remote_file_path="/path/to/target.csv",
+        ... )
+        >>> os.fspath(remote_file)
+        '/path/to/target.csv'
+        >>> connection.path_exists("/path/to/target.csv")
+        True
+        >>> remote_file.stat().st_size  # in bytes
+        1024
+        >>> os.stat("/path/to/source.csv").st_size  # same as source
+        1024
         """
 
     @abstractmethod
@@ -566,6 +614,8 @@ class BaseFileConnection(BaseConnection):
         r"""
         Returns string content of a file at specific path. |support_hooks|
 
+        .. versionadded:: 0.8.0
+
         Parameters
         ----------
         path : str or :obj:`os.PathLike`
@@ -589,10 +639,8 @@ class BaseFileConnection(BaseConnection):
         Examples
         --------
 
-        .. code:: python
-
-            content = connection.read_text("/path/to/dir/file.csv")
-            assert content == "some;header\n1;2"
+        >>> connection.read_text("/path/to/dir/file.csv")
+        'some;header\n1;2'
         """
 
     @abstractmethod
@@ -600,6 +648,8 @@ class BaseFileConnection(BaseConnection):
         """
         Returns binary content of a file at specific path. |support_hooks|
 
+        .. versionadded:: 0.8.0
+
         Parameters
         ----------
         path : str or :obj:`os.PathLike`
@@ -620,10 +670,8 @@ class BaseFileConnection(BaseConnection):
         Examples
         --------
 
-        .. code:: python
-
-            content = connection.read_bytes("/path/to/dir/file.csv")
-            assert content == b"0xdeadbeef"
+        >>> connection.read_bytes("/path/to/dir/file.csv")
+        b'0xdeadbeef'
         """
 
     @abstractmethod
@@ -640,6 +688,8 @@ class BaseFileConnection(BaseConnection):
 
             If file already exists, its content will be replaced.
 
+        .. versionadded:: 0.8.0
+
         Parameters
         ----------
         path : str or :obj:`os.PathLike`
@@ -666,10 +716,11 @@ class BaseFileConnection(BaseConnection):
         Examples
         --------
 
-        .. code:: python
-
-            file_path = connection.write_text("/path/to/dir/file.csv", "some;header\n1;2")
-            assert file_path.stat.st_size > 0
+        >>> file_path = connection.write_text("/path/to/dir/file.csv", "some;header\n1;2")
+        >>> os.fspath(file_path)
+        '/path/to/dir/file.csv'
+        >>> file_path.stat().st_size  # in bytes
+        1024
         """
 
     @abstractmethod
@@ -681,6 +732,8 @@ class BaseFileConnection(BaseConnection):
 
             If file already exists, its content will be replaced.
 
+        .. versionadded:: 0.8.0
+
         Parameters
         ----------
         path : str or :obj:`os.PathLike`
@@ -704,10 +757,11 @@ class BaseFileConnection(BaseConnection):
         Examples
         --------
 
-        .. code:: python
-
-            file_path = connection.write_bytes("/path/to/dir/file.csv", b"0xdeadbeef")
-            assert file_path.stat.st_size > 0
+        >>> file_path = connection.write_bytes("/path/to/dir/file.csv", b"0xdeadbeef")
+        >>> os.fspath(file_path)
+        '/path/to/dir/file.csv'
+        >>> file_path.stat().st_size  # in bytes
+        1024
         """
 
     @property
