@@ -1,14 +1,11 @@
+# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
+# SPDX-License-Identifier: Apache-2.0
 from __future__ import annotations
 
-from etl_entities import Table
-
-from onetl.base import BaseDBConnection
-
 
 class SupportNameWithSchemaOnly:
-    @classmethod
-    def validate_name(cls, connection: BaseDBConnection, value: Table) -> Table:
-        if value.name.count(".") != 1:
+    def validate_name(self, value: str) -> str:
+        if "." not in value:
             raise ValueError(
                 f"Name should be passed in `schema.name` format, got '{value}'",
             )
