@@ -1,17 +1,5 @@
-#  Copyright 2023 MTS (Mobile Telesystems)
-#
-#  Licensed under the Apache License, Version 2.0 (the "License");
-#  you may not use this file except in compliance with the License.
-#  You may obtain a copy of the License at
-#
-#      http://www.apache.org/licenses/LICENSE-2.0
-#
-#  Unless required by applicable law or agreed to in writing, software
-#  distributed under the License is distributed on an "AS IS" BASIS,
-#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-#  See the License for the specific language governing permissions and
-#  limitations under the License.
-
+# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
+# SPDX-License-Identifier: Apache-2.0
 from __future__ import annotations
 
 import logging
@@ -19,15 +7,26 @@ import os
 import textwrap
 import warnings
 from typing import TYPE_CHECKING, Any, ClassVar
+from urllib.parse import quote, urlencode, urlparse, urlunparse
 
 from etl_entities.instance import Host
-from pydantic import validator
 
-from onetl._internal import get_sql_query
 from onetl._util.classproperty import classproperty
+from onetl.connection.db_connection.jdbc_connection.options import JDBCReadOptions
+
+try:
+    from pydantic.v1 import SecretStr, validator
+except (ImportError, AttributeError):
+    from pydantic import validator, SecretStr  # type: ignore[no-redef, assignment]
+
 from onetl._util.java import try_import_java_class
 from onetl._util.scala import get_default_scala_version
-from onetl._util.spark import get_executor_total_cores, get_spark_version
+from onetl._util.spark import (
+    get_client_info,
+    get_executor_total_cores,
+    get_spark_version,
+    override_job_description,
+)
 from onetl._util.version import Version
 from onetl.connection.db_connection.db_connection import DBConnection
 from onetl.connection.db_connection.greenplum.connection_limit import (
@@ -35,15 +34,24 @@ from onetl.connection.db_connection.greenplum.connection_limit import (
 )
 from onetl.connection.db_connection.greenplum.dialect import GreenplumDialect
 from onetl.connection.db_connection.greenplum.options import (
+    GreenplumExecuteOptions,
+    GreenplumFetchOptions,
     GreenplumReadOptions,
+    GreenplumSQLOptions,
     GreenplumTableExistBehavior,
     GreenplumWriteOptions,
 )
 from onetl.connection.db_connection.jdbc_mixin import JDBCMixin
-from onetl.connection.db_connection.jdbc_mixin.options import JDBCOptions
+from onetl.connection.db_connection.jdbc_mixin.options import (
+    JDBCExecuteOptions,
+    JDBCFetchOptions,
+)
+from onetl.connection.db_connection.jdbc_mixin.options import (
+    JDBCOptions as JDBCMixinOptions,
+)
 from onetl.exception import MISSING_JVM_CLASS_MSG, TooManyParallelJobsError
 from onetl.hooks import slot, support_hooks
-from onetl.hwm import Statement
+from onetl.hwm import Window
 from onetl.impl import GenericOptions
 from onetl.log import log_lines, log_with_indent
 
@@ -69,20 +77,22 @@ class GreenplumExtra(GenericOptions):
 
     class Config:
         extra = "allow"
-        prohibited_options = JDBCOptions.Config.prohibited_options
+        prohibited_options = JDBCMixinOptions.Config.prohibited_options
 
 
 @support_hooks
-class Greenplum(JDBCMixin, DBConnection):
+class Greenplum(JDBCMixin, DBConnection):  # noqa: WPS338
     """Greenplum connection. |support_hooks|
 
-    Based on package ``io.pivotal:greenplum-spark:2.1.4``
-    (`Pivotal connector for Spark <https://network.tanzu.vmware.com/products/vmware-greenplum#/releases/1341690/file_groups/14993>`_).
+    Based on package ``io.pivotal:greenplum-spark:2.2.0``
+    (`VMware Greenplum connector for Spark <https://docs.vmware.com/en/VMware-Greenplum-Connector-for-Apache-Spark/index.html>`_).
 
-    .. warning::
+    .. seealso::
 
         Before using this connector please take into account :ref:`greenplum-prerequisites`
 
+    .. versionadded:: 0.5.0
+
     Parameters
     ----------
     host : str
@@ -111,47 +121,37 @@ class Greenplum(JDBCMixin, DBConnection):
         For example: ``{"tcpKeepAlive": "true", "server.port": "50000-65535"}``
 
         Supported options are:
-            * All `Postgres JDBC driver properties <https://github.com/pgjdbc/pgjdbc#connection-properties>`_
-            * Properties from `Greenplum connector for Spark documentation <https://docs.vmware.com/en/VMware-Tanzu-Greenplum-Connector-for-Apache-Spark/2.1/tanzu-greenplum-connector-spark/GUID-options.html>`_ page, but only starting with ``server.`` or ``pool.``
+            * All `Postgres JDBC driver properties <https://jdbc.postgresql.org/documentation/use/>`_
+            * Properties from `Greenplum connector for Spark documentation <https://docs.vmware.com/en/VMware-Greenplum-Connector-for-Apache-Spark/2.3/greenplum-connector-spark/options.html>`_ page, but only starting with ``server.`` or ``pool.``
 
     Examples
     --------
 
-    Greenplum connection initialization
+    Create and check Greenplum connection:
 
     .. code:: python
 
         from onetl.connection import Greenplum
         from pyspark.sql import SparkSession
 
-        # Please ask your DevOps and Greenplum admin what port range
-        # on Spark side can be used to accept requests from Greenplum segments
-
-        extra = {
-            "server.port": "49152-65535",
-        }
-
         # Create Spark session with Greenplum connector loaded
-        # See Prerequisites page for more details
         maven_packages = Greenplum.get_packages(spark_version="3.2")
         spark = (
             SparkSession.builder.appName("spark-app-name")
             .config("spark.jars.packages", ",".join(maven_packages))
+            .config("spark.executor.allowSparkContext", "true")
+            # IMPORTANT!!!
+            # Set number of executors according to "Prerequisites" -> "Number of executors"
             .config("spark.dynamicAllocation.maxExecutors", 10)
             .config("spark.executor.cores", 1)
             .getOrCreate()
         )
 
         # IMPORTANT!!!
-        # Each job on the Spark executor make its own connection to Greenplum master node,
-        # so we need to limit number of connections to avoid opening too many of them.
-        #
-        # Table size ~20Gb requires about 10 executors * cores,
-        # ~50Gb requires ~ 20 executors * cores,
-        # 100Gb+ requires 30 executors * cores.
-        #
-        # Cores number can be increased, but executors count should be reduced
-        # to keep the same number of executors * cores.
+        # Set port range of executors according to "Prerequisites" -> "Network ports"
+        extra = {
+            "server.port": "41000-42000",
+        }
 
         # Create connection
         greenplum = Greenplum(
@@ -161,36 +161,53 @@ class Greenplum(JDBCMixin, DBConnection):
             database="target_database",
             extra=extra,
             spark=spark,
-        )
+        ).check()
     """
 
     host: Host
+    user: str
+    password: SecretStr
     database: str
     port: int = 5432
     extra: GreenplumExtra = GreenplumExtra()
 
-    Extra = GreenplumExtra
-    Dialect = GreenplumDialect
     ReadOptions = GreenplumReadOptions
     WriteOptions = GreenplumWriteOptions
+    SQLOptions = GreenplumSQLOptions
+    FetchOptions = GreenplumFetchOptions
+    ExecuteOptions = GreenplumExecuteOptions
+    JDBCOptions = JDBCMixinOptions
+
+    Extra = GreenplumExtra
+    Dialect = GreenplumDialect
 
     DRIVER: ClassVar[str] = "org.postgresql.Driver"
     CONNECTIONS_WARNING_LIMIT: ClassVar[int] = 31
     CONNECTIONS_EXCEPTION_LIMIT: ClassVar[int] = 100
 
+    _CHECK_QUERY: ClassVar[str] = "SELECT 1"
+    # any small table with always present in db, and which any user can access
+    # https://www.greenplumdba.com/pg-catalog-tables-and-views
+    _CHECK_DUMMY_TABLE: ClassVar[str] = "pg_catalog.gp_id"
+    _CHECK_DUMMY_PARTITION_COLUMN: ClassVar[str] = "dbid"
+
     @slot
     @classmethod
     def get_packages(
         cls,
+        *,
         scala_version: str | None = None,
         spark_version: str | None = None,
+        package_version: str | None = None,
     ) -> list[str]:
         """
         Get package names to be downloaded by Spark. |support_hooks|
 
         .. warning::
 
-            You should pass at least one parameter.
+            You should pass either ``scala_version`` or ``spark_version``.
+
+        .. versionadded:: 0.9.0
 
         Parameters
         ----------
@@ -204,6 +221,11 @@ class Greenplum(JDBCMixin, DBConnection):
 
             Used only if ``scala_version=None``.
 
+        package_version : str, optional, default ``2.2.0``
+            Package version in format ``major.minor.patch``
+
+            .. versionadded:: 0.10.1
+
         Examples
         --------
 
@@ -211,63 +233,89 @@ class Greenplum(JDBCMixin, DBConnection):
 
             from onetl.connection import Greenplum
 
-            Greenplum.get_packages(scala_version="2.11")
-            Greenplum.get_packages(spark_version="3.2")
+            Greenplum.get_packages(scala_version="2.12")
+            Greenplum.get_packages(spark_version="3.2", package_version="2.3.0")
 
         """
 
         # Connector version is fixed, so we can perform checks for Scala/Spark version
+        if package_version:
+            package_ver = Version(package_version)
+        else:
+            package_ver = Version("2.2.0")
+
         if scala_version:
-            scala_ver = Version.parse(scala_version)
+            scala_ver = Version(scala_version).min_digits(2)
         elif spark_version:
-            spark_ver = Version.parse(spark_version)
-            if spark_ver.digits(2) > (3, 2) or spark_ver.digits(2) < (2, 3):
-                raise ValueError(f"Spark version must be 2.3.x - 3.2.x, got {spark_ver}")
+            spark_ver = Version(spark_version).min_digits(2)
+            if spark_ver >= Version("3.3"):
+                raise ValueError(f"Spark version must be 3.2.x or less, got {spark_ver}")
             scala_ver = get_default_scala_version(spark_ver)
         else:
             raise ValueError("You should pass either `scala_version` or `spark_version`")
 
-        if scala_ver.digits(2) < (2, 11) or scala_ver.digits(2) > (2, 12):
-            raise ValueError(f"Scala version must be 2.11 - 2.12, got {scala_ver}")
-
-        return [f"io.pivotal:greenplum-spark_{scala_ver.digits(2)}:2.1.4"]
-
-    @classproperty
-    def package_spark_2_3(cls) -> str:
-        """Get package name to be downloaded by Spark 2.3."""
-        msg = "`Greenplum.package_2_3` will be removed in 1.0.0, use `Greenplum.get_packages(spark_version='2.3')` instead"
-        warnings.warn(msg, UserWarning, stacklevel=3)
-        return "io.pivotal:greenplum-spark_2.11:2.1.4"
-
-    @classproperty
-    def package_spark_2_4(cls) -> str:
-        """Get package name to be downloaded by Spark 2.4."""
-        msg = "`Greenplum.package_2_4` will be removed in 1.0.0, use `Greenplum.get_packages(spark_version='2.4')` instead"
-        warnings.warn(msg, UserWarning, stacklevel=3)
-        return "io.pivotal:greenplum-spark_2.11:2.1.4"
+        return [f"io.pivotal:greenplum-spark_{scala_ver.format('{0}.{1}')}:{package_ver}"]
 
     @classproperty
     def package_spark_3_2(cls) -> str:
         """Get package name to be downloaded by Spark 3.2."""
         msg = "`Greenplum.package_3_2` will be removed in 1.0.0, use `Greenplum.get_packages(spark_version='3.2')` instead"
         warnings.warn(msg, UserWarning, stacklevel=3)
-        return "io.pivotal:greenplum-spark_2.12:2.1.4"
+        return "io.pivotal:greenplum-spark_2.12:2.2.0"
 
     @property
     def instance_url(self) -> str:
         return f"{self.__class__.__name__.lower()}://{self.host}:{self.port}/{self.database}"
 
+    def __str__(self):
+        return f"{self.__class__.__name__}[{self.host}:{self.port}/{self.database}]"
+
     @property
     def jdbc_url(self) -> str:
-        extra = {
+        return f"jdbc:postgresql://{self.host}:{self.port}/{self.database}"
+
+    @property
+    def jdbc_custom_params(self) -> dict:
+        result = {
             key: value
             for key, value in self.extra.dict(by_alias=True).items()
             if not (key.startswith("server.") or key.startswith("pool."))
         }
-        extra["ApplicationName"] = extra.get("ApplicationName", self.spark.sparkContext.appName)
+        # https://www.postgresql.org/docs/current/runtime-config-logging.html#GUC-APPLICATION-NAME
+        result["ApplicationName"] = result.get("ApplicationName", get_client_info(self.spark, limit=64))
+        return result
 
-        parameters = "&".join(f"{k}={v}" for k, v in sorted(extra.items()))
-        return f"jdbc:postgresql://{self.host}:{self.port}/{self.database}?{parameters}".rstrip("?")
+    @property
+    def jdbc_params(self) -> dict:
+        result = super().jdbc_params
+        result.update(self.jdbc_custom_params)
+        return result
+
+    @slot
+    def check(self):
+        log.info("|%s| Checking connection availability...", self.__class__.__name__)
+        self._log_parameters()  # type: ignore
+
+        log.debug("|%s| Executing SQL query:", self.__class__.__name__)
+        log_lines(log, self._CHECK_QUERY, level=logging.DEBUG)
+
+        try:
+            with override_job_description(self.spark, f"{self}.check()"):
+                self._query_optional_on_driver(self._CHECK_QUERY, self.FetchOptions(fetchsize=1))
+
+                read_options = self._get_connector_params(self._CHECK_DUMMY_TABLE)
+                # do not require gp_segment_id column in table
+                read_options["partitions"] = 1
+                read_options["partitionColumn"] = self._CHECK_DUMMY_PARTITION_COLUMN
+                df = self.spark.read.format("greenplum").options(**read_options).load()
+                df.take(1)
+
+            log.info("|%s| Connection is available.", self.__class__.__name__)
+        except Exception as e:
+            log.exception("|%s| Connection is unavailable", self.__class__.__name__)
+            raise RuntimeError("Connection is unavailable") from e
+
+        return self
 
     @slot
     def read_source_as_df(
@@ -277,25 +325,29 @@ class Greenplum(JDBCMixin, DBConnection):
         hint: str | None = None,
         where: str | None = None,
         df_schema: StructType | None = None,
-        start_from: Statement | None = None,
-        end_at: Statement | None = None,
+        window: Window | None = None,
+        limit: int | None = None,
         options: GreenplumReadOptions | None = None,
     ) -> DataFrame:
         read_options = self.ReadOptions.parse(options).dict(by_alias=True, exclude_none=True)
         log.info("|%s| Executing SQL query (on executor):", self.__class__.__name__)
-        where = self.Dialect._condition_assembler(condition=where, start_from=start_from, end_at=end_at)
-        query = get_sql_query(table=source, columns=columns, where=where)
-        log_lines(log, query)
+        where = self.dialect.apply_window(where, window)
+        fake_query_for_log = self.dialect.get_sql_query(table=source, columns=columns, where=where, limit=limit)
+        log_lines(log, fake_query_for_log)
 
-        df = self.spark.read.format("greenplum").options(**self._connector_params(source), **read_options).load()
+        df = self.spark.read.format("greenplum").options(**self._get_connector_params(source), **read_options).load()
         self._check_expected_jobs_number(df, action="read")
 
         if where:
-            df = df.filter(where)
+            for item in where:
+                df = df.filter(item)
 
         if columns:
             df = df.selectExpr(*columns)
 
+        if limit is not None:
+            df = df.limit(limit)
+
         log.info("|Spark| DataFrame successfully created from SQL statement ")
         return df
 
@@ -318,7 +370,7 @@ class Greenplum(JDBCMixin, DBConnection):
             else write_options.if_exists.value
         )
         df.write.format("greenplum").options(
-            **self._connector_params(target),
+            **self._get_connector_params(target),
             **options_dict,
         ).mode(mode).save()
 
@@ -329,61 +381,60 @@ class Greenplum(JDBCMixin, DBConnection):
         self,
         source: str,
         columns: list[str] | None = None,
-        options: JDBCOptions | None = None,
+        options: JDBCReadOptions | None = None,
     ) -> StructType:
+        log.info("|%s| Detected dialect: '%s'", self.__class__.__name__, self._get_spark_dialect_class_name())
         log.info("|%s| Fetching schema of table %r ...", self.__class__.__name__, source)
 
-        query = get_sql_query(source, columns=columns, where="1=0", compact=True)
-        jdbc_options = self.JDBCOptions.parse(options).copy(update={"fetchsize": 0})
+        query = self.dialect.get_sql_query(source, columns=columns, limit=0, compact=True)
+        jdbc_options = self.ReadOptions.parse(options).copy(update={"fetchsize": 0})
 
         log.debug("|%s| Executing SQL query (on driver):", self.__class__.__name__)
         log_lines(log, query, level=logging.DEBUG)
 
-        df = self._query_on_driver(query, jdbc_options)
+        df = self._query_on_driver(query, self.FetchOptions.parse(jdbc_options.dict()))
         log.info("|%s| Schema fetched.", self.__class__.__name__)
 
         return df.schema
 
     @slot
-    def get_min_max_bounds(
+    def get_min_max_values(
         self,
         source: str,
-        column: str,
-        expression: str | None = None,
-        hint: str | None = None,
-        where: str | None = None,
-        options: JDBCOptions | None = None,
+        window: Window,
+        hint: Any | None = None,
+        where: Any | None = None,
+        options: JDBCReadOptions | None = None,
     ) -> tuple[Any, Any]:
-        log.info("|%s| Getting min and max values for column %r ...", self.__class__.__name__, column)
+        log.info("|%s| Getting min and max values for %r ...", self.__class__.__name__, window.expression)
+        jdbc_options = self.ReadOptions.parse(options).copy(update={"fetchsize": 1})
 
-        jdbc_options = self.JDBCOptions.parse(options).copy(update={"fetchsize": 1})
-
-        query = get_sql_query(
+        query = self.dialect.get_sql_query(
             table=source,
             columns=[
-                self.Dialect._expression_with_alias(
-                    self.Dialect._get_min_value_sql(expression or column),
-                    self.Dialect._escape_column("min"),
+                self.dialect.aliased(
+                    self.dialect.get_min_value(window.expression),
+                    self.dialect.escape_column("min"),
                 ),
-                self.Dialect._expression_with_alias(
-                    self.Dialect._get_max_value_sql(expression or column),
-                    self.Dialect._escape_column("max"),
+                self.dialect.aliased(
+                    self.dialect.get_max_value(window.expression),
+                    self.dialect.escape_column("max"),
                 ),
             ],
-            where=where,
+            where=self.dialect.apply_window(where, window),
         )
 
         log.info("|%s| Executing SQL query (on driver):", self.__class__.__name__)
         log_lines(log, query)
 
-        df = self._query_on_driver(query, jdbc_options)
+        df = self._query_on_driver(query, self.FetchOptions.parse(jdbc_options.dict()))
         row = df.collect()[0]
         min_value = row["min"]
         max_value = row["max"]
 
         log.info("|%s| Received values:", self.__class__.__name__)
-        log_with_indent(log, "MIN(%r) = %r", column, min_value)
-        log_with_indent(log, "MAX(%r) = %r", column, max_value)
+        log_with_indent(log, "MIN(%s) = %r", window.expression, min_value)
+        log_with_indent(log, "MAX(%s) = %r", window.expression, max_value)
 
         return min_value, max_value
 
@@ -394,43 +445,52 @@ class Greenplum(JDBCMixin, DBConnection):
         try:
             try_import_java_class(spark, java_class)
         except Exception as e:
-            spark_version = get_spark_version(spark).digits(2)
+            spark_version = get_spark_version(spark).format("{0}.{1}")
             msg = MISSING_JVM_CLASS_MSG.format(
                 java_class=java_class,
                 package_source=cls.__name__,
                 args=f"spark_version='{spark_version}'",
             )
-            if log.isEnabledFor(logging.DEBUG):
-                log.debug("Missing Java class", exc_info=e, stack_info=True)
             raise ValueError(msg) from e
         return spark
 
-    def _connector_params(
+    def _get_connector_params(
         self,
         table: str,
     ) -> dict:
         schema, table_name = table.split(".")  # noqa: WPS414
         extra = self.extra.dict(by_alias=True, exclude_none=True)
-        extra = {key: value for key, value in extra.items() if key.startswith("server.") or key.startswith("pool.")}
+        greenplum_connector_options = {
+            key: value for key, value in extra.items() if key.startswith("server.") or key.startswith("pool.")
+        }
+
+        # Greenplum connector requires all JDBC params to be passed via JDBC URL:
+        # https://docs.vmware.com/en/VMware-Greenplum-Connector-for-Apache-Spark/2.3/greenplum-connector-spark/using_the_connector.html#specifying-session-parameters
+        parsed_jdbc_url = urlparse(self.jdbc_url)
+        sorted_jdbc_params = [(k, v) for k, v in sorted(self.jdbc_custom_params.items(), key=lambda x: x[0].lower())]
+        jdbc_url_query = urlencode(sorted_jdbc_params, quote_via=quote)
+        jdbc_url = urlunparse(parsed_jdbc_url._replace(query=jdbc_url_query))
+
         return {
             "driver": self.DRIVER,
-            "url": self.jdbc_url,
+            "url": jdbc_url,
             "user": self.user,
             "password": self.password.get_secret_value(),
             "dbschema": schema,
             "dbtable": table_name,
-            **extra,
+            **greenplum_connector_options,
         }
 
-    def _options_to_connection_properties(self, options: JDBCOptions):
-        # See https://github.com/pgjdbc/pgjdbc/pull/1252
-        # Since 42.2.9 Postgres JDBC Driver added new option readOnlyMode=transaction
-        # Which is not a desired behavior, because `.fetch()` method should always be read-only
+    def _get_jdbc_connection(self, options: JDBCFetchOptions | JDBCExecuteOptions, read_only: bool):
+        if read_only:
+            # To properly support pgbouncer, we have to create connection with readOnly option set.
+            # See https://github.com/pgjdbc/pgjdbc/issues/848
+            options = options.copy(update={"readOnly": True})
 
-        if not getattr(options, "readOnlyMode", None):
-            options = options.copy(update={"readOnlyMode": "always"})
-
-        return super()._options_to_connection_properties(options)
+        connection_properties = self._options_to_connection_properties(options)
+        driver_manager = self.spark._jvm.java.sql.DriverManager  # type: ignore
+        # avoid calling .setReadOnly(True) here
+        return driver_manager.getConnection(self.jdbc_url, connection_properties)
 
     def _get_server_setting(self, name: str) -> Any:
         query = f"""
@@ -441,7 +501,7 @@ class Greenplum(JDBCMixin, DBConnection):
         log.debug("|%s| Executing SQL query (on driver):")
         log_lines(log, query, level=logging.DEBUG)
 
-        df = self._query_on_driver(query, self.JDBCOptions())
+        df = self._query_on_driver(query, self.FetchOptions())
         result = df.collect()
 
         log.debug(
@@ -462,7 +522,7 @@ class Greenplum(JDBCMixin, DBConnection):
         log.debug("|%s| Executing SQL query (on driver):")
         log_lines(log, query, level=logging.DEBUG)
 
-        df = self._query_on_driver(query, self.JDBCOptions())
+        df = self._query_on_driver(query, self.FetchOptions())
         result = df.collect()
 
         log.debug(
