@@ -1,28 +1,22 @@
-#  Copyright 2023 MTS (Mobile Telesystems)
-#
-#  Licensed under the Apache License, Version 2.0 (the "License");
-#  you may not use this file except in compliance with the License.
-#  You may obtain a copy of the License at
-#
-#      http://www.apache.org/licenses/LICENSE-2.0
-#
-#  Unless required by applicable law or agreed to in writing, software
-#  distributed under the License is distributed on an "AS IS" BASIS,
-#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-#  See the License for the specific language governing permissions and
-#  limitations under the License.
-
+# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
+# SPDX-License-Identifier: Apache-2.0
 from __future__ import annotations
 
 import logging
 from textwrap import dedent
-from typing import TYPE_CHECKING, Any, ClassVar, Iterable, Tuple
+from typing import TYPE_CHECKING, Any, ClassVar, Iterable
 
 from etl_entities.instance import Cluster
-from pydantic import validator
 
-from onetl._internal import clear_statement, get_sql_query
-from onetl._util.spark import inject_spark_param
+try:
+    from pydantic.v1 import validator
+except (ImportError, AttributeError):
+    from pydantic import validator  # type: ignore[no-redef, assignment]
+
+from onetl._metrics.recorder import SparkMetricsRecorder
+from onetl._util.spark import inject_spark_param, override_job_description
+from onetl._util.sql import clear_statement
+from onetl.base import BaseWritableFileFormat
 from onetl.connection.db_connection.db_connection import DBConnection
 from onetl.connection.db_connection.hive.dialect import HiveDialect
 from onetl.connection.db_connection.hive.options import (
@@ -31,8 +25,9 @@ from onetl.connection.db_connection.hive.options import (
     HiveWriteOptions,
 )
 from onetl.connection.db_connection.hive.slots import HiveSlots
+from onetl.file.format.file_format import ReadWriteFileFormat, WriteOnlyFileFormat
 from onetl.hooks import slot, support_hooks
-from onetl.hwm import Statement
+from onetl.hwm import Window
 from onetl.log import log_lines, log_with_indent
 
 if TYPE_CHECKING:
@@ -47,94 +42,66 @@ log = logging.getLogger(__name__)
 class Hive(DBConnection):
     """Spark connection with Hive MetaStore support. |support_hooks|
 
-    You don't need a Hive server to use this connector.
+    .. seealso::
 
-    .. dropdown:: Version compatibility
+        Before using this connector please take into account :ref:`hive-prerequisites`
 
-        * Hive metastore version: 0.12 - 3.1.2 (may require to add proper .jar file explicitly)
-        * Spark versions: 2.3.x - 3.4.x
-        * Java versions: 8 - 20
-
-    .. warning::
-
-        To use Hive connector you should have PySpark installed (or injected to ``sys.path``)
-        BEFORE creating the connector instance.
-
-        You can install PySpark as follows:
-
-        .. code:: bash
-
-            pip install onetl[spark]  # latest PySpark version
-
-            # or
-            pip install onetl pyspark=3.4.1  # pass specific PySpark version
-
-        See :ref:`spark-install` instruction for more details.
-
-    .. warning::
-
-        This connector requires some additional configuration files to be present (``hive-site.xml`` and so on),
-        as well as .jar files with Hive MetaStore client.
-
-        See `Spark Hive Tables documentation <https://spark.apache.org/docs/latest/sql-data-sources-hive-tables.html>`_
-        and `this guide <https://dataedo.com/docs/apache-spark-hive-metastore>`_ for more details.
-
-    .. note::
-
-        Most of Hadoop instances use Kerberos authentication. In this case, you should call ``kinit``
-        **BEFORE** starting Spark session to generate Kerberos ticket. See :ref:`kerberos-install`.
-
-        In case of creating session with ``"spark.master": "yarn"``, you should also pass some additional options
-        to Spark session, allowing executors to generate their own Kerberos tickets to access HDFS.
-        See `Spark security documentation <https://spark.apache.org/docs/latest/security.html#kerberos>`_
-        for more details.
+    .. versionadded:: 0.1.0
 
     Parameters
     ----------
     cluster : str
         Cluster name. Used for HWM and lineage.
 
+        .. versionadded:: 0.7.0
+
     spark : :obj:`pyspark.sql.SparkSession`
         Spark session with Hive metastore support enabled
 
     Examples
     --------
 
-    Hive connection initialization
+    .. tabs::
 
-    .. code:: python
+        .. tab:: Create Hive connection with Kerberos auth
 
-        from onetl.connection import Hive
-        from pyspark.sql import SparkSession
+            Execute ``kinit`` consome command before creating Spark Session
 
-        # Create Spark session
-        spark = SparkSession.builder.appName("spark-app-name").enableHiveSupport().getOrCreate()
+            .. code:: bash
 
-        # Create connection
-        hive = Hive(cluster="rnd-dwh", spark=spark).check()
+                $ kinit -kt /path/to/keytab user
 
-    Hive connection initialization with Kerberos support
+            .. code:: python
 
-    .. code:: python
+                from onetl.connection import Hive
+                from pyspark.sql import SparkSession
 
-        from onetl.connection import Hive
-        from pyspark.sql import SparkSession
+                # Create Spark session
+                # Use names "spark.yarn.access.hadoopFileSystems", "spark.yarn.principal"
+                # and "spark.yarn.keytab" for Spark 2
 
-        # Create Spark session
-        # Use names "spark.yarn.access.hadoopFileSystems", "spark.yarn.principal"
-        # and "spark.yarn.keytab" for Spark 2
+                spark = (
+                    SparkSession.builder.appName("spark-app-name")
+                    .option("spark.kerberos.access.hadoopFileSystems", "hdfs://cluster.name.node:8020")
+                    .option("spark.kerberos.principal", "user")
+                    .option("spark.kerberos.keytab", "/path/to/keytab")
+                    .enableHiveSupport()
+                    .getOrCreate()
+                )
 
-        spark = (
-            SparkSession.builder.appName("spark-app-name")
-            .option("spark.kerberos.access.hadoopFileSystems", "hdfs://cluster.name.node:8020")
-            .option("spark.kerberos.principal", "user")
-            .option("spark.kerberos.keytab", "/path/to/keytab")
-            .enableHiveSupport()
-            .getOrCreate()
-        )
+                # Create connection
+                hive = Hive(cluster="rnd-dwh", spark=spark).check()
 
-        # Create connection
-        hive = Hive(cluster="rnd-dwh", spark=spark).check()
+        .. code-tab:: py Create Hive connection with anonymous auth
+
+            from onetl.connection import Hive
+            from pyspark.sql import SparkSession
+
+            # Create Spark session
+            spark = SparkSession.builder.appName("spark-app-name").enableHiveSupport().getOrCreate()
+
+            # Create connection
+            hive = Hive(cluster="rnd-dwh", spark=spark).check()
     """
 
     cluster: Cluster
@@ -146,7 +113,7 @@ class Hive(DBConnection):
     # TODO: remove in v1.0.0
     slots = HiveSlots
 
-    _CHECK_QUERY: ClassVar[str] = "SELECT 1"
+    _CHECK_QUERY: ClassVar[str] = "SHOW DATABASES"
 
     @slot
     @classmethod
@@ -159,6 +126,8 @@ class Hive(DBConnection):
             Can be used only if there are some hooks bound to
             :obj:`Slots.get_current_cluster <onetl.connection.db_connection.hive.slots.HiveSlots.get_current_cluster>` slot.
 
+        .. versionadded:: 0.7.0
+
         Parameters
         ----------
         spark : :obj:`pyspark.sql.SparkSession`
@@ -193,6 +162,9 @@ class Hive(DBConnection):
     def instance_url(self) -> str:
         return self.cluster
 
+    def __str__(self):
+        return f"{self.__class__.__name__}[{self.cluster}]"
+
     @slot
     def check(self):
         log.debug("|%s| Detecting current cluster...", self.__class__.__name__)
@@ -207,7 +179,8 @@ class Hive(DBConnection):
         log_lines(log, self._CHECK_QUERY, level=logging.DEBUG)
 
         try:
-            self._execute_sql(self._CHECK_QUERY)
+            with override_job_description(self.spark, f"{self}.check()"):
+                self._execute_sql(self._CHECK_QUERY).take(1)
             log.info("|%s| Connection is available.", self.__class__.__name__)
         except Exception as e:
             log.exception("|%s| Connection is unavailable", self.__class__.__name__)
@@ -225,32 +198,19 @@ class Hive(DBConnection):
 
         Same as ``spark.sql(query)``.
 
+        .. versionadded:: 0.2.0
+
         Parameters
         ----------
         query : str
 
-            SQL query to be executed, like:
-
-            * ``SELECT ... FROM ...``
-            * ``WITH ... AS (...) SELECT ... FROM ...``
-            * ``SHOW ...`` queries are also supported, like ``SHOW TABLES``
+            SQL query to be executed.
 
         Returns
         -------
         df : pyspark.sql.dataframe.DataFrame
 
             Spark dataframe
-
-        Examples
-        --------
-
-        Read data from Hive table:
-
-        .. code:: python
-
-            connection = Hive(cluster="rnd-dwh", spark=spark)
-
-            df = connection.sql("SELECT * FROM mytable")
         """
 
         query = clear_statement(query)
@@ -258,8 +218,30 @@ class Hive(DBConnection):
         log.info("|%s| Executing SQL query:", self.__class__.__name__)
         log_lines(log, query)
 
-        df = self._execute_sql(query)
-        log.info("|Spark| DataFrame successfully created from SQL statement")
+        with SparkMetricsRecorder(self.spark) as recorder:
+            try:
+                with override_job_description(self.spark, f"{self}.sql()"):
+                    df = self._execute_sql(query)
+            except Exception:
+                log.error("|%s| Query failed", self.__class__.__name__)
+
+                metrics = recorder.metrics()
+                if log.isEnabledFor(logging.DEBUG) and not metrics.is_empty:
+                    # as SparkListener results are not guaranteed to be received in time,
+                    # some metrics may be missing. To avoid confusion, log only in debug, and with a notice
+                    log.info("|%s| Recorded metrics (some values may be missing!):", self.__class__.__name__)
+                    log_lines(log, str(metrics), level=logging.DEBUG)
+                raise
+
+            log.info("|Spark| DataFrame successfully created from SQL statement")
+
+            metrics = recorder.metrics()
+            if log.isEnabledFor(logging.DEBUG) and not metrics.is_empty:
+                # as SparkListener results are not guaranteed to be received in time,
+                # some metrics may be missing. To avoid confusion, log only in debug, and with a notice
+                log.info("|%s| Recorded metrics (some values may be missing!):", self.__class__.__name__)
+                log_lines(log, str(metrics), level=logging.DEBUG)
+
         return df
 
     @slot
@@ -270,47 +252,13 @@ class Hive(DBConnection):
         """
         Execute DDL or DML statement. |support_hooks|
 
+        .. versionadded:: 0.2.0
+
         Parameters
         ----------
         statement : str
 
-            Statement to be executed, like:
-
-            DML statements:
-
-            * ``INSERT INTO target_table SELECT * FROM source_table``
-            * ``TRUNCATE TABLE mytable``
-
-            DDL statements:
-
-            * ``CREATE TABLE mytable (...)``
-            * ``ALTER TABLE mytable ...``
-            * ``DROP TABLE mytable``
-            * ``MSCK REPAIR TABLE mytable``
-
-            The exact list of supported statements depends on Hive version,
-            for example some new versions support ``CREATE FUNCTION`` syntax.
-
-        Examples
-        --------
-
-        Create table:
-
-        .. code:: python
-
-            connection = Hive(cluster="rnd-dwh", spark=spark)
-
-            connection.execute(
-                "CREATE TABLE mytable (id NUMBER, data VARCHAR) PARTITIONED BY (date DATE)"
-            )
-
-        Drop table partition:
-
-        .. code:: python
-
-            connection = Hive(cluster="rnd-dwh", spark=spark)
-
-            connection.execute("ALTER TABLE mytable DROP PARTITION(date='2023-02-01')")
+            Statement to be executed.
         """
 
         statement = clear_statement(statement)
@@ -318,8 +266,28 @@ class Hive(DBConnection):
         log.info("|%s| Executing statement:", self.__class__.__name__)
         log_lines(log, statement)
 
-        self._execute_sql(statement).collect()
-        log.info("|%s| Call succeeded", self.__class__.__name__)
+        with SparkMetricsRecorder(self.spark) as recorder:
+            try:
+                with override_job_description(self.spark, f"{self}.execute()"):
+                    self._execute_sql(statement).collect()
+            except Exception:
+                log.error("|%s| Execution failed", self.__class__.__name__)
+                metrics = recorder.metrics()
+                if log.isEnabledFor(logging.DEBUG) and not metrics.is_empty:
+                    # as SparkListener results are not guaranteed to be received in time,
+                    # some metrics may be missing. To avoid confusion, log only in debug, and with a notice
+                    log.info("|%s| Recorded metrics (some values may be missing!):", self.__class__.__name__)
+                    log_lines(log, str(metrics), level=logging.DEBUG)
+                raise
+
+            log.info("|%s| Execution succeeded", self.__class__.__name__)
+
+            metrics = recorder.metrics()
+            if log.isEnabledFor(logging.DEBUG) and not metrics.is_empty:
+                # as SparkListener results are not guaranteed to be received in time,
+                # some metrics may be missing. To avoid confusion, log only in debug, and with a notice
+                log.info("|%s| Recorded metrics (some values may be missing!):", self.__class__.__name__)
+                log_lines(log, str(metrics), level=logging.DEBUG)
 
     @slot
     def write_df_to_target(
@@ -330,23 +298,26 @@ class Hive(DBConnection):
     ) -> None:
         write_options = self.WriteOptions.parse(options)
 
-        try:
-            self.get_df_schema(target)
-            table_exists = True
-
-            log.info("|%s| Table %r already exists", self.__class__.__name__, target)
-        except Exception:
-            table_exists = False
-
         # https://stackoverflow.com/a/72747050
-        if table_exists and write_options.if_exists != HiveTableExistBehavior.REPLACE_ENTIRE_TABLE:
-            # using saveAsTable on existing table does not handle
-            # spark.sql.sources.partitionOverwriteMode=dynamic, so using insertInto instead.
-            self._insert_into(df, target, options)
-        else:
+        if not self._target_exist(target) or write_options.if_exists == HiveTableExistBehavior.REPLACE_ENTIRE_TABLE:
             # if someone needs to recreate the entire table using new set of options, like partitionBy or bucketBy,
             # if_exists="replace_entire_table" should be used
             self._save_as_table(df, target, options)
+            return
+
+        if write_options.if_exists == HiveTableExistBehavior.ERROR:
+            raise ValueError("Operation stopped due to Hive.WriteOptions(if_exists='error')")
+
+        if write_options.if_exists == HiveTableExistBehavior.IGNORE:
+            log.info(
+                "|%s| Skip writing to existing table because of Hive.WriteOptions(if_exists='ignore')",
+                self.__class__.__name__,
+            )
+            return
+
+        # using saveAsTable on existing table does not handle
+        # spark.sql.sources.partitionOverwriteMode=dynamic, so using insertInto instead.
+        self._insert_into(df, target, options)
 
     @slot
     def read_source_as_df(
@@ -356,18 +327,18 @@ class Hive(DBConnection):
         hint: str | None = None,
         where: str | None = None,
         df_schema: StructType | None = None,
-        start_from: Statement | None = None,
-        end_at: Statement | None = None,
+        window: Window | None = None,
+        limit: int | None = None,
     ) -> DataFrame:
-        where = self.Dialect._condition_assembler(condition=where, start_from=start_from, end_at=end_at)
-        sql_text = get_sql_query(
+        query = self.dialect.get_sql_query(
             table=source,
             columns=columns,
-            where=where,
+            where=self.dialect.apply_window(where, window),
             hint=hint,
+            limit=limit,
         )
 
-        return self.sql(sql_text)
+        return self.sql(query)
 
     @slot
     def get_df_schema(
@@ -375,8 +346,8 @@ class Hive(DBConnection):
         source: str,
         columns: list[str] | None = None,
     ) -> StructType:
-        log.info("|%s| Fetching schema of table table %r ...", self.__class__.__name__, source)
-        query = get_sql_query(source, columns=columns, where="1=0", compact=True)
+        log.info("|%s| Fetching schema of table %r ...", self.__class__.__name__, source)
+        query = self.dialect.get_sql_query(source, columns=columns, where=0, compact=True)
 
         log.debug("|%s| Executing SQL query:", self.__class__.__name__)
         log_lines(log, query, level=logging.DEBUG)
@@ -386,43 +357,42 @@ class Hive(DBConnection):
         return df.schema
 
     @slot
-    def get_min_max_bounds(
+    def get_min_max_values(
         self,
         source: str,
-        column: str,
-        expression: str | None = None,
-        hint: str | None = None,
-        where: str | None = None,
-    ) -> Tuple[Any, Any]:
-        log.info("|%s| Getting min and max values for column %r ...", self.__class__.__name__, column)
+        window: Window,
+        hint: Any | None = None,
+        where: Any | None = None,
+    ) -> tuple[Any, Any]:
+        log.info("|%s| Getting min and max values for expression %r ...", self.__class__.__name__, window.expression)
 
-        sql_text = get_sql_query(
+        query = self.dialect.get_sql_query(
             table=source,
             columns=[
-                self.Dialect._expression_with_alias(
-                    self.Dialect._get_min_value_sql(expression or column),
-                    self.Dialect._escape_column("min"),
+                self.dialect.aliased(
+                    self.dialect.get_min_value(window.expression),
+                    self.dialect.escape_column("min"),
                 ),
-                self.Dialect._expression_with_alias(
-                    self.Dialect._get_max_value_sql(expression or column),
-                    self.Dialect._escape_column("max"),
+                self.dialect.aliased(
+                    self.dialect.get_max_value(window.expression),
+                    self.dialect.escape_column("max"),
                 ),
             ],
-            where=where,
+            where=self.dialect.apply_window(where, window),
             hint=hint,
         )
 
-        log.debug("|%s| Executing SQL query:", self.__class__.__name__)
-        log_lines(log, sql_text, level=logging.DEBUG)
+        log.info("|%s| Executing SQL query:", self.__class__.__name__)
+        log_lines(log, query)
 
-        df = self._execute_sql(sql_text)
+        df = self._execute_sql(query)
         row = df.collect()[0]
         min_value = row["min"]
         max_value = row["max"]
 
         log.info("|%s| Received values:", self.__class__.__name__)
-        log_with_indent(log, "MIN(%s) = %r", column, min_value)
-        log_with_indent(log, "MAX(%s) = %r", column, max_value)
+        log_with_indent(log, "MIN(%s) = %r", window.expression, min_value)
+        log_with_indent(log, "MAX(%s) = %r", window.expression, max_value)
 
         return min_value, max_value
 
@@ -491,6 +461,29 @@ class Hive(DBConnection):
 
         return sorted(df_columns, key=lambda column: table_columns_normalized.index(column.casefold()))
 
+    def _target_exist(self, name: str) -> bool:
+        from pyspark.sql.functions import col, lower
+
+        log.info("|%s| Checking if table %r exists ...", self.__class__.__name__, name)
+
+        # Do not use SELECT * FROM table, because it may fail if users have no permissions,
+        # or Hive Metastore is overloaded.
+        # Also we ignore VIEW's as they are not insertable.
+        schema, table = name.rsplit(".", maxsplit=1)
+        query = f"SHOW TABLES IN {schema} LIKE '{table}'"
+
+        log.debug("|%s| Executing SQL query:", self.__class__.__name__)
+        log_lines(log, query, level=logging.DEBUG)
+
+        # Spark normalizes table names to lowercase, so we do the same
+        df = self._execute_sql(query).where(lower(col("tableName")) == table.lower())
+        if df.take(1):
+            log.info("|%s| Table %r exists.", self.__class__.__name__, name)
+            return True
+
+        log.info("|%s| Table %r does not exist.", self.__class__.__name__, name)
+        return False
+
     def _insert_into(
         self,
         df: DataFrame,
@@ -498,11 +491,11 @@ class Hive(DBConnection):
         options: HiveWriteOptions | dict | None = None,
     ) -> None:
         write_options = self.WriteOptions.parse(options)
-
-        unsupported_options = write_options.dict(by_alias=True, exclude_unset=True, exclude={"if_exists"})
+        unsupported_options = self._format_write_options(write_options)
         if unsupported_options:
             log.warning(
-                "|%s| Options %r are not supported while inserting into existing table, ignoring",
+                "|%s| User-specified options %r are ignored while inserting into existing table. "
+                "Using only table parameters from Hive metastore",
                 self.__class__.__name__,
                 unsupported_options,
             )
@@ -511,18 +504,32 @@ class Hive(DBConnection):
         # So we should sort columns according their order in the existing table
         # instead of using order from the dataframe
         columns = self._sort_df_columns_like_table(table, df.columns)
-        writer = df.select(*columns).write
+        if columns != df.columns:
+            df = df.select(*columns)
 
-        # Writer option "partitionOverwriteMode" was added to Spark only in 2.4.0
-        # so using a workaround with patching Spark config and then setting up the previous value
+        # Explicit option "partitionOverwriteMode" requires DataFrameWriterV2 (TODO).
+        # Using a workaround with patching Spark config and then setting up the previous value
+        log.info("|%s| Inserting data into existing table %r ...", self.__class__.__name__, table)
         with inject_spark_param(self.spark.conf, PARTITION_OVERWRITE_MODE_PARAM, "dynamic"):
             overwrite = write_options.if_exists != HiveTableExistBehavior.APPEND
 
-            log.info("|%s| Inserting data into existing table %r ...", self.__class__.__name__, table)
-            writer.insertInto(table, overwrite=overwrite)
+            df.write.insertInto(table, overwrite=overwrite)
 
         log.info("|%s| Data is successfully inserted into table %r.", self.__class__.__name__, table)
 
+    def _format_write_options(self, write_options: HiveWriteOptions) -> dict:
+        options_dict = write_options.dict(
+            by_alias=True,
+            exclude_unset=True,
+            exclude={"if_exists"},
+        )
+
+        if isinstance(write_options.format, (WriteOnlyFileFormat, ReadWriteFileFormat)):
+            options_dict["format"] = write_options.format.name
+            options_dict.update(write_options.format.dict(exclude={"name"}, exclude_none=True))
+
+        return options_dict
+
     def _save_as_table(
         self,
         df: DataFrame,
@@ -532,17 +539,25 @@ class Hive(DBConnection):
         write_options = self.WriteOptions.parse(options)
 
         writer = df.write
-        for method, value in write_options.dict(by_alias=True, exclude_none=True, exclude={"if_exists"}).items():
-            # <value> is the arguments that will be passed to the <method>
-            # format orc, parquet methods and format simultaneously
+        for method, value in write_options.dict(  # noqa: WPS352
+            by_alias=True,
+            exclude_none=True,
+            exclude={"if_exists", "format"},
+        ).items():
             if hasattr(writer, method):
                 if isinstance(value, Iterable) and not isinstance(value, str):
-                    writer = getattr(writer, method)(*value)  # noqa: WPS220
+                    writer = getattr(writer, method)(*value)
                 else:
-                    writer = getattr(writer, method)(value)  # noqa: WPS220
+                    writer = getattr(writer, method)(value)
             else:
                 writer = writer.option(method, value)
 
+        # deserialize passed ORC(), Parquet(), CSV(), etc. file formats
+        if isinstance(write_options.format, BaseWritableFileFormat):
+            writer = write_options.format.apply_to_writer(writer)
+        elif isinstance(write_options.format, str):
+            writer = writer.format(write_options.format)
+
         mode = "append" if write_options.if_exists == HiveTableExistBehavior.APPEND else "overwrite"
 
         log.info("|%s| Saving data to a table %r ...", self.__class__.__name__, table)
