@@ -1,30 +1,27 @@
-#  Copyright 2023 MTS (Mobile Telesystems)
-#
-#  Licensed under the Apache License, Version 2.0 (the "License");
-#  you may not use this file except in compliance with the License.
-#  You may obtain a copy of the License at
-#
-#      http://www.apache.org/licenses/LICENSE-2.0
-#
-#  Unless required by applicable law or agreed to in writing, software
-#  distributed under the License is distributed on an "AS IS" BASIS,
-#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-#  See the License for the specific language governing permissions and
-#  limitations under the License.
-
+# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
+# SPDX-License-Identifier: Apache-2.0
 from __future__ import annotations
 
+import json
 import logging
 from contextlib import closing
 from typing import TYPE_CHECKING, Any, List, Optional
 
 from etl_entities.instance import Cluster
-from pydantic import root_validator, validator
 
-from onetl._internal import stringify
+try:
+    from pydantic.v1 import root_validator, validator
+except (ImportError, AttributeError):
+    from pydantic import root_validator, validator  # type: ignore[no-redef, assignment]
+
 from onetl._util.java import try_import_java_class
 from onetl._util.scala import get_default_scala_version
-from onetl._util.spark import get_spark_version
+from onetl._util.spark import (
+    get_client_info,
+    get_spark_version,
+    override_job_description,
+    stringify,
+)
 from onetl._util.version import Version
 from onetl.connection.db_connection.db_connection import DBConnection
 from onetl.connection.db_connection.kafka.dialect import KafkaDialect
@@ -46,7 +43,7 @@ from onetl.connection.db_connection.kafka.options import (
 from onetl.connection.db_connection.kafka.slots import KafkaSlots
 from onetl.exception import MISSING_JVM_CLASS_MSG, TargetAlreadyExistsError
 from onetl.hooks import slot, support_hooks
-from onetl.hwm import Statement
+from onetl.hwm.window import Window
 from onetl.log import log_collection, log_with_indent
 
 if TYPE_CHECKING:
@@ -64,14 +61,15 @@ class Kafka(DBConnection):
     Based on `official Kafka Source For Spark
     <https://spark.apache.org/docs/latest/structured-streaming-kafka-integration.html>`_.
 
+    .. seealso::
+
+        Before using this connector please take into account :ref:`kafka-prerequisites`
+
     .. note::
 
-        This connector is for batch download from kafka and not streaming.
+        This connector is for **batch** ETL processes, not streaming.
 
-    .. dropdown:: Version compatibility
-
-        * Apache Kafka versions: 0.10 or higher
-        * Spark versions: 2.4.x - 3.4.x
+    .. versionadded:: 0.9.0
 
     Parameters
     ----------
@@ -113,116 +111,97 @@ class Kafka(DBConnection):
             Options that populated from connection
             attributes (like ``bootstrap.servers``, ``sasl.*``, ``ssl.*``) are not allowed to override.
 
-        .. note::
-
-            At current version Kafka connection doesn't support batch strategies.
-
     Examples
     --------
 
-    Connect to Kafka using ``PLAINTEXT`` protocol and without any auth mechanism (anonymous user, default):
+    .. tabs::
 
-    .. code:: python
+        .. code-tab:: py Create Kafka connection with ``PLAINTEXT`` protocol and ``SCRAM-SHA-256`` auth
 
-        from onetl.connection import Kafka
-        from pyspark.sql import SparkSession
+            from onetl.connection import Kafka
+            from pyspark.sql import SparkSession
 
-        # Create Spark session with Kafka connector loaded
-        maven_packages = Kafka.get_packages(spark_version="3.2.4")
-        spark = (
-            SparkSession.builder.appName("spark-app-name")
-            .config("spark.jars.packages", ",".join(maven_packages))
-            .getOrCreate()
-        )
+            # Create Spark session with Kafka connector loaded
+            maven_packages = Kafka.get_packages(spark_version="3.5.6")
+            exclude_packages = Kafka.get_exclude_packages()
+            spark = (
+                SparkSession.builder.appName("spark-app-name")
+                .config("spark.jars.packages", ",".join(maven_packages))
+                .config("spark.jars.excludes", ",".join(exclude_packages))
+                .getOrCreate()
+            )
 
-        # Create connection
-        kafka = Kafka(
-            addresses=["mybroker:9092", "anotherbroker:9092"],
-            cluster="my-cluster",
-            spark=spark,
-        )
+            # Create connection
+            kafka = Kafka(
+                addresses=["mybroker:9092", "anotherbroker:9092"],
+                cluster="my-cluster",
+                auth=Kafka.ScramAuth(
+                    user="me",
+                    password="abc",
+                    digest="SHA-256",
+                ),
+                spark=spark,
+            ).check()
 
-    Connect to Kafka using ``PLAINTEXT`` protocol and basic (``PLAIN``) auth mechanism:
+        .. code-tab:: py Create Kafka connection with ``PLAINTEXT`` protocol and Kerberos (``GSSAPI``) auth
 
-    .. code:: python
+            # Create Spark session with Kafka connector loaded
+            ...
 
-        # Create Spark session with Kafka connector loaded
-        ...
+            # Create connection
+            kafka = Kafka(
+                addresses=["mybroker:9092", "anotherbroker:9092"],
+                cluster="my-cluster",
+                auth=Kafka.KerberosAuth(
+                    principal="me@example.com",
+                    keytab="/path/to/keytab",
+                    deploy_keytab=True,
+                ),
+                spark=spark,
+            ).check()
 
-        # Create connection
-        kafka = Kafka(
-            addresses=["mybroker:9092", "anotherbroker:9092"],
-            cluster="my-cluster",
-            auth=Kafka.BasicAuth(
-                user="me",
-                password="password",
-            ),
-            spark=spark,
-        )
+        .. code-tab:: py Create Kafka connection with ``SASL_SSL`` protocol and ``SCRAM-SHA-512`` auth
 
-    Connect to Kafka using ``PLAINTEXT`` protocol and Kerberos (``GSSAPI``) auth mechanism:
+            from pathlib import Path
 
-    .. code:: python
+            # Create Spark session with Kafka connector loaded
+            ...
 
-        # Create Spark session with Kafka connector loaded
-        ...
+            # Create connection
+            kafka = Kafka(
+                addresses=["mybroker:9092", "anotherbroker:9092"],
+                cluster="my-cluster",
+                protocol=Kafka.SSLProtocol(
+                    # read client certificate and private key from file
+                    keystore_type="PEM",
+                    keystore_certificate_chain=Path("path/to/user.crt").read_text(),
+                    keystore_key=Path("path/to/user.key").read_text(),
+                    # read server public certificate from file
+                    truststore_type="PEM",
+                    truststore_certificates=Path("/path/to/server.crt").read_text(),
+                ),
+                auth=Kafka.ScramAuth(
+                    user="me",
+                    password="abc",
+                    digest="SHA-512",
+                ),
+                spark=spark,
+            ).check()
 
-        # Create connection
-        kafka = Kafka(
-            addresses=["mybroker:9092", "anotherbroker:9092"],
-            cluster="my-cluster",
-            auth=Kafka.KerberosAuth(
-                principal="me@example.com",
-                keytab="/path/to/keytab",
-                deploy_keytab=True,
-            ),
-            spark=spark,
-        )
+        .. code-tab:: py Create Kafka connection with extra options
 
-    Connect to Kafka using ``SASL_SSL`` protocol and ``SCRAM-SHA-512`` auth mechanism:
+            # Create Spark session with Kafka connector loaded
+            ...
 
-    .. code:: python
-
-        from pathlib import Path
-
-        # Create Spark session with Kafka connector loaded
-        ...
-
-        # Create connection
-        kafka = Kafka(
-            addresses=["mybroker:9092", "anotherbroker:9092"],
-            cluster="my-cluster",
-            protocol=Kafka.SSLProtocol(
-                keystore_type="PEM",
-                keystore_certificate_chain=Path("path/to/user.crt").read_text(),
-                keystore_key=Path("path/to/user.key").read_text(),
-                truststore_type="PEM",
-                truststore_certificates=Path("/path/to/server.crt").read_text(),
-            ),
-            auth=Kafka.ScramAuth(
-                user="me",
-                password="abc",
-                digest="SHA-512",
-            ),
-            spark=spark,
-        )
-
-    Connect to Kafka with extra options:
-
-    .. code:: python
-
-        # Create Spark session with Kafka connector loaded
-        ...
-
-        # Create connection
-        kafka = Kafka(
-            addresses=["mybroker:9092", "anotherbroker:9092"],
-            cluster="my-cluster",
-            protocol=...,
-            auth=...,
-            extra={"max.request.size": 1000000},
-            spark=spark,
-        )
+            # Create connection
+            kafka = Kafka(
+                addresses=["mybroker:9092", "anotherbroker:9092"],
+                cluster="my-cluster",
+                protocol=...,
+                auth=...,
+                extra={"max.request.size": 1024 * 1024},  # <--
+                spark=spark,
+            ).check()
 
     """
 
@@ -249,7 +228,14 @@ class Kafka(DBConnection):
         self._log_parameters()
 
         try:
-            self._get_topics()
+            with override_job_description(self.spark, f"{self}.check()"):
+                self._get_topics()
+
+                read_options = {f"kafka.{key}": value for key, value in self._get_connection_properties().items()}
+                # We need to read just any topic allowed to check if Kafka is alive
+                read_options["subscribePattern"] = ".*"
+                self.spark.read.format("kafka").options(**read_options).load().take(1)
+
             log.info("|%s| Connection is available.", self.__class__.__name__)
         except Exception as e:
             log.exception("|%s| Connection is unavailable", self.__class__.__name__)
@@ -257,15 +243,15 @@ class Kafka(DBConnection):
         return self
 
     @slot
-    def read_source_as_df(
+    def read_source_as_df(  # noqa:  WPS231
         self,
         source: str,
         columns: list[str] | None = None,
         hint: Any | None = None,
         where: Any | None = None,
         df_schema: StructType | None = None,
-        start_from: Statement | None = None,
-        end_at: Statement | None = None,
+        window: Window | None = None,
+        limit: int | None = None,
         options: KafkaReadOptions = KafkaReadOptions(),  # noqa: B008, WPS404
     ) -> DataFrame:
         log.info("|%s| Reading data from topic %r", self.__class__.__name__, source)
@@ -275,7 +261,30 @@ class Kafka(DBConnection):
         result_options = {f"kafka.{key}": value for key, value in self._get_connection_properties().items()}
         result_options.update(options.dict(by_alias=True, exclude_none=True))
         result_options["subscribe"] = source
+
+        if window and window.expression == "offset":
+            # the 'including' flag in window values are relevant for batch strategies which are not
+            # supported by Kafka, therefore we always get offsets including border values
+            starting_offsets = dict(window.start_from.value) if window.start_from.value else {}
+            ending_offsets = dict(window.stop_at.value) if window.stop_at.value else {}
+
+            # when the Kafka topic's number of partitions has increased during incremental processing,
+            # new partitions, which are present in ending_offsets but not in
+            # starting_offsets, are assigned a default offset (0 in this case).
+            for partition in ending_offsets:
+                if partition not in starting_offsets:
+                    starting_offsets[partition] = 0
+
+            if starting_offsets:
+                result_options["startingOffsets"] = json.dumps({source: starting_offsets})
+            if ending_offsets:
+                result_options["endingOffsets"] = json.dumps({source: ending_offsets})
+
         df = self.spark.read.format("kafka").options(**result_options).load()
+
+        if limit is not None:
+            df = df.limit(limit)
+
         log.info("|%s| Dataframe is successfully created.", self.__class__.__name__)
         return df
 
@@ -303,12 +312,6 @@ class Kafka(DBConnection):
         if not options.include_headers and "headers" in df.columns:
             raise ValueError("Cannot write 'headers' column with kafka.WriteOptions(include_headers=False)")
 
-        spark_version = get_spark_version(self.spark)
-        if options.include_headers and spark_version.major < 3:
-            raise ValueError(
-                f"kafka.WriteOptions(include_headers=True) requires Spark 3.x, got {spark_version}",
-            )
-
         if "topic" in df.columns:
             log.warning("The 'topic' column in the DataFrame will be overridden with value %r", target)
 
@@ -316,7 +319,7 @@ class Kafka(DBConnection):
         write_options.update(options.dict(by_alias=True, exclude_none=True, exclude={"if_exists"}))
         write_options["topic"] = target
 
-        # As of Apache Spark version 3.4.1, the mode 'error' is not functioning as expected.
+        # As of Apache Spark version 3.5.6, the mode 'error' is not functioning as expected.
         # This issue has been reported and can be tracked at:
         # https://issues.apache.org/jira/browse/SPARK-44774
         mode = options.if_exists
@@ -345,31 +348,31 @@ class Kafka(DBConnection):
             TimestampType,
         )
 
-        schema = StructType(
-            [
-                StructField("key", BinaryType(), nullable=True),
-                StructField("value", BinaryType(), nullable=False),
-                StructField("topic", StringType(), nullable=True),
-                StructField("partition", IntegerType(), nullable=True),
-                StructField("offset", LongType(), nullable=True),
-                StructField("timestamp", TimestampType(), nullable=True),
-                StructField("timestampType", IntegerType(), nullable=True),
-                StructField(
-                    "headers",
-                    ArrayType(
-                        StructType(
-                            [
-                                StructField("key", StringType(), nullable=True),
-                                StructField("value", BinaryType(), nullable=True),
-                            ],
-                        ),
+        all_fields = [
+            StructField("key", BinaryType(), nullable=True),
+            StructField("value", BinaryType(), nullable=False),
+            StructField("topic", StringType(), nullable=True),
+            StructField("partition", IntegerType(), nullable=True),
+            StructField("offset", LongType(), nullable=True),
+            StructField("timestamp", TimestampType(), nullable=True),
+            StructField("timestampType", IntegerType(), nullable=True),
+            StructField(
+                "headers",
+                ArrayType(
+                    StructType(
+                        [
+                            StructField("key", StringType(), nullable=True),
+                            StructField("value", BinaryType(), nullable=True),
+                        ],
                     ),
-                    nullable=True,
                 ),
-            ],
-        )
+                nullable=True,
+            ),
+        ]
 
-        return schema  # noqa:  WPS331
+        filtered_fields = [field for field in all_fields if columns is None or field.name in columns]
+
+        return StructType(filtered_fields)
 
     @slot
     @classmethod
@@ -381,6 +384,9 @@ class Kafka(DBConnection):
         """
         Get package names to be downloaded by Spark. |support_hooks|
 
+        See `Maven package index <https://mvnrepository.com/artifact/org.apache.spark/spark-sql-kafka-0-10>`_
+        for all available packages.
+
         Parameters
         ----------
         spark_version : str
@@ -398,23 +404,49 @@ class Kafka(DBConnection):
 
             from onetl.connection import Kafka
 
-            Kafka.get_packages(spark_version="3.2.4")
-            Kafka.get_packages(spark_version="3.2.4", scala_version="2.13")
+            Kafka.get_packages(spark_version="3.5.6")
+            Kafka.get_packages(spark_version="3.5.6", scala_version="2.12")
 
         """
 
-        # Connector version is same as Spark, do not perform any additional checks
-        spark_ver = Version.parse(spark_version)
-        if spark_ver < (2, 4):
-            # Kafka connector for Spark 2.3 is build with Kafka client 0.10.0.1 which does not support
-            # passing `sasl.jaas.config` option. It is supported only in 0.10.2.0,
-            # see https://issues.apache.org/jira/browse/KAFKA-4259
-            # Old client requires generating JAAS file and placing it to filesystem, which is not secure.
-            raise ValueError(f"Spark version must be at least 2.4, got {spark_ver}")
-
-        scala_ver = Version.parse(scala_version) if scala_version else get_default_scala_version(spark_ver)
+        spark_ver = Version(spark_version).min_digits(3)
+        scala_ver = Version(scala_version).min_digits(2) if scala_version else get_default_scala_version(spark_ver)
         return [
-            f"org.apache.spark:spark-sql-kafka-0-10_{scala_ver.digits(2)}:{spark_ver.digits(3)}",
+            f"org.apache.spark:spark-sql-kafka-0-10_{scala_ver.format('{0}.{1}')}:{spark_ver.format('{0}.{1}.{2}')}",
+        ]
+
+    @slot
+    @classmethod
+    def get_exclude_packages(cls) -> list[str]:
+        """
+        Get package names to be excluded by Spark. |support_hooks|
+
+        .. versionadded:: 0.13.0
+
+        Examples
+        --------
+
+        .. code:: python
+
+            from onetl.connection import Kafka
+
+            Kafka.get_exclude_packages()
+
+        """
+
+        return [
+            # already a part of Spark bundle
+            "org.apache.hadoop:hadoop-client-api",
+            "org.apache.hadoop:hadoop-client-runtime",
+            "com.fasterxml.jackson.core:jackson-annotations",
+            "com.fasterxml.jackson.core:jackson-core",
+            "com.fasterxml.jackson.core:jackson-databind",
+            "com.google.code.findbugs:jsr305",
+            "commons-codec:commons-codec",
+            "commons-logging:commons-logging",
+            "org.lz4:lz4-java",
+            "org.slf4j:slf4j-api",
+            "org.xerial.snappy:snappy-java",
         ]
 
     def __enter__(self):
@@ -458,10 +490,75 @@ class Kafka(DBConnection):
             self.auth.cleanup(self)
         return self
 
+    # Do not all __del__ with calling .close(), like other connections,
+    # because this can influence dataframes created by this connection.
+    # For example, .close() deletes local keytab copy.
+
+    @slot
+    def get_min_max_values(
+        self,
+        source: str,
+        window: Window,
+        hint: Any | None = None,
+        where: Any | None = None,
+        options: KafkaReadOptions | dict | None = None,
+    ) -> tuple[dict[int, int], dict[int, int]]:
+        log.info("|%s| Getting min and max offset values for topic %r ...", self.__class__.__name__, source)
+
+        consumer = self._get_java_consumer()
+        with closing(consumer):
+            # https://kafka.apache.org/22/javadoc/org/apache/kafka/clients/consumer/KafkaConsumer.html#partitionsFor-java.lang.String-
+            partition_infos = consumer.partitionsFor(source)
+
+            jvm = self.spark._jvm  # type: ignore[attr-defined]
+            topic_partitions = [
+                jvm.org.apache.kafka.common.TopicPartition(source, p.partition())  # type: ignore[union-attr]
+                for p in partition_infos
+            ]
+
+            # https://kafka.apache.org/22/javadoc/org/apache/kafka/clients/consumer/KafkaConsumer.html#beginningOffsets-java.util.Collection-
+            beginning_offsets = consumer.beginningOffsets(topic_partitions)
+
+            # https://kafka.apache.org/22/javadoc/org/apache/kafka/clients/consumer/KafkaConsumer.html#endOffsets-java.util.Collection-
+            end_offsets = consumer.endOffsets(topic_partitions)
+
+            min_offsets = {}
+            max_offsets = {}
+            for tp in topic_partitions:
+                partition_id = tp.partition()
+                begin_offset = beginning_offsets[tp]
+                end_offset = end_offsets[tp]
+
+                if window.start_from and window.start_from.is_set():
+                    window_start = window.start_from.value.get(partition_id, begin_offset)
+                    begin_offset = max(window_start, begin_offset)
+                if window.stop_at and window.stop_at.is_set():
+                    window_stop = window.stop_at.value.get(partition_id, end_offset)
+                    end_offset = min(window_stop, end_offset)
+
+                min_offsets[partition_id] = begin_offset
+                max_offsets[partition_id] = end_offset
+
+        log.info("|%s| Received min and max offset values for each partition.", self.__class__.__name__)
+        partitions = sorted(set(min_offsets.keys() | max_offsets.keys()))
+        for partition_id in partitions:
+            log.debug(
+                "|%s| Partition %d: Min Offset = %d, Max Offset = %d",
+                self.__class__.__name__,
+                partition_id,
+                min_offsets.get(partition_id),
+                max_offsets.get(partition_id),
+            )
+
+        return min_offsets, max_offsets
+
     @property
     def instance_url(self):
         return "kafka://" + self.cluster
 
+    def __str__(self):
+        return f"{self.__class__.__name__}[{self.cluster}]"
+
     @root_validator(pre=True)
     def _get_addresses_by_cluster(cls, values):
         cluster = values.get("cluster")
@@ -508,14 +605,6 @@ class Kafka(DBConnection):
 
         return validated_addresses
 
-    @validator("spark")
-    def _check_spark_version(cls, spark):
-        spark_version = get_spark_version(spark)
-        if spark_version < (2, 4):
-            raise ValueError(f"Spark version must be at least 2.4, got {spark_version}")
-
-        return spark
-
     @validator("spark")
     def _check_java_class_imported(cls, spark):
         java_class = "org.apache.spark.sql.kafka010.KafkaSourceProvider"
@@ -523,14 +612,12 @@ class Kafka(DBConnection):
         try:
             try_import_java_class(spark, java_class)
         except Exception as e:
-            spark_version = get_spark_version(spark).digits(2)
+            spark_version = get_spark_version(spark).format("{0}.{1}")
             msg = MISSING_JVM_CLASS_MSG.format(
                 java_class=java_class,
                 package_source=cls.__name__,
                 args=f"spark_version='{spark_version}'",
             )
-            if log.isEnabledFor(logging.DEBUG):
-                log.debug("Missing Java class", exc_info=e, stack_info=True)
             raise ValueError(msg) from e
         return spark
 
@@ -541,7 +628,7 @@ class Kafka(DBConnection):
         if self.auth:
             result.update(self.auth.get_options(self))
 
-        result["client.id"] = self.spark.sparkContext.appName  # type: ignore[assignment]
+        result["client.id"] = result.get("client.id", get_client_info(self.spark))
         return stringify(result)
 
     def _get_java_consumer(self):
@@ -559,7 +646,7 @@ class Kafka(DBConnection):
         return consumer_class(connection_properties)
 
     def _get_topics(self, timeout: int = 10) -> set[str]:
-        jvm = self.spark._jvm
+        jvm = self.spark._jvm  # type: ignore[attr-defined]
         # Maybe we should not pass explicit timeout at all,
         # and instead use default.api.timeout.ms which is configurable via self.extra.
         # Think about this next time if someone see issues in real use
