@@ -1,27 +1,21 @@
-#  Copyright 2023 MTS (Mobile Telesystems)
-#
-#  Licensed under the Apache License, Version 2.0 (the "License");
-#  you may not use this file except in compliance with the License.
-#  You may obtain a copy of the License at
-#
-#      http://www.apache.org/licenses/LICENSE-2.0
-#
-#  Unless required by applicable law or agreed to in writing, software
-#  distributed under the License is distributed on an "AS IS" BASIS,
-#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-#  See the License for the specific language governing permissions and
-#  limitations under the License.
-
+# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
+# SPDX-License-Identifier: Apache-2.0
 from __future__ import annotations
 
 import os
 import threading
 from abc import abstractmethod
+from contextlib import suppress
 from logging import getLogger
 from typing import Any, Iterable, Iterator
 
 from humanize import naturalsize
 
+try:
+    from pydantic.v1 import PrivateAttr
+except (ImportError, AttributeError):
+    from pydantic import PrivateAttr  # type: ignore[no-redef, assignment]
+
 from onetl.base import (
     BaseFileConnection,
     BaseFileFilter,
@@ -52,7 +46,7 @@ log = getLogger(__name__)
 
 @support_hooks
 class FileConnection(BaseFileConnection, FrozenModel):
-    _clients_cache: Any = None
+    _clients_cache: Any = PrivateAttr(default=None)
 
     @property
     def client(self):
@@ -72,8 +66,10 @@ class FileConnection(BaseFileConnection, FrozenModel):
             if client and not self._is_client_closed(client):
                 return client
         except AttributeError:
-            self._clients_cache.client = self._get_client()
-            return self._clients_cache.client
+            pass
+
+        self._clients_cache.client = self._get_client()
+        return self._clients_cache.client
 
     @slot
     def close(self):
@@ -96,7 +92,6 @@ class FileConnection(BaseFileConnection, FrozenModel):
         .. code:: python
 
             content = connection.list_dir("/mydir")
-            assert content
             connection.close()
 
             # or
@@ -112,8 +107,14 @@ class FileConnection(BaseFileConnection, FrozenModel):
         except AttributeError:
             return self
 
-        self._close_client(client)
-        del self._clients_cache.client
+        with suppress(Exception):
+            # exceptions while closing client should be ignored
+            self._close_client(client)
+
+        with suppress(Exception):
+            # .close() could be called from destructor, and modifying self is not allowed here
+            del self._clients_cache.client
+
         return self
 
     def __enter__(self):
@@ -122,6 +123,10 @@ class FileConnection(BaseFileConnection, FrozenModel):
     def __exit__(self, _exc_type, _exc_value, _traceback):
         self.close()
 
+    def __del__(self):  # noqa: WPS603
+        # If current object is collected by GC, close opened connection
+        self.close()
+
     @slot
     def check(self):
         log.info("|%s| Checking connection availability...", self.__class__.__name__)
@@ -167,24 +172,25 @@ class FileConnection(BaseFileConnection, FrozenModel):
     def resolve_dir(self, path: os.PathLike | str) -> RemoteDirectory:
         is_dir = self.is_dir(path)
         stat = self.get_stat(path)
+        remote_path = RemotePath(path)
 
         if not is_dir:
-            raise NotADirectoryError(
-                f"{path_repr(RemoteFile(path, stats=stat))} is not a directory",
-            )
+            remote_file = RemoteFile(path=remote_path, stats=stat)
+            raise NotADirectoryError(f"{path_repr(remote_file)} is not a directory")
 
-        return RemoteDirectory(path=path, stats=stat)
+        return RemoteDirectory(path=remote_path, stats=stat)
 
     @slot
     def resolve_file(self, path: os.PathLike | str) -> RemoteFile:
         is_file = self.is_file(path)
         stat = self.get_stat(path)
-        remote_path = RemoteFile(path=path, stats=stat)
+        remote_path = RemotePath(path)
 
         if not is_file:
-            raise NotAFileError(f"{path_repr(remote_path)} is not a file")
+            remote_directory = RemoteDirectory(path=remote_path, stats=stat)
+            raise NotAFileError(f"{path_repr(remote_directory)} is not a file")
 
-        return remote_path
+        return RemoteFile(path=remote_path, stats=stat)
 
     @slot
     def read_text(self, path: os.PathLike | str, encoding: str = "utf-8", **kwargs) -> str:
@@ -410,20 +416,22 @@ class FileConnection(BaseFileConnection, FrozenModel):
         limits = reset_limits(limits or [])
 
         for entry in self._scan_entries(remote_dir):
+            if limits_reached(limits):
+                break
+
             name = self._extract_name_from_entry(entry)
             stat = self._extract_stat_from_entry(remote_dir, entry)
 
             if self._is_dir_entry(remote_dir, entry):
-                path = RemoteDirectory(path=name, stats=stat)
+                path = RemoteDirectory(path=remote_dir / name, stats=stat)
+            elif self._is_file_entry(remote_dir, entry):
+                path = RemoteFile(path=remote_dir / name, stats=stat)
             else:
-                path = RemoteFile(path=name, stats=stat)
+                continue
 
-            if match_all_filters(path, filters):
+            if match_all_filters(path, filters) and not limits_stop_at(path, limits):
                 result.append(path)
 
-            if limits_stop_at(path, limits):
-                break
-
         return result
 
     @slot
@@ -455,12 +463,15 @@ class FileConnection(BaseFileConnection, FrozenModel):
             return False
 
         directory_info = path_repr(self.resolve_dir(remote_dir))
-        if self.list_dir(remote_dir) and not recursive:
-            raise DirectoryNotEmptyError(
-                "|%s| Cannot delete non-empty directory %s",
-                self.__class__.__name__,
-                directory_info,
-            )
+        if not recursive:
+            # self.list_dir may return large list
+            # self._scan_entries return an iterator, which have to be iterated at least once
+            for _entry in self._scan_entries(remote_dir):  # noqa: WPS122, WPS328
+                raise DirectoryNotEmptyError(
+                    "|%s| Cannot delete non-empty directory %s",
+                    self.__class__.__name__,
+                    directory_info,
+                )
 
         log.debug("|%s| Directory to remove: %s", self.__class__.__name__, directory_info)
         if recursive:
@@ -486,29 +497,27 @@ class FileConnection(BaseFileConnection, FrozenModel):
         dirs, files = [], []
 
         for entry in self._scan_entries(root):
+            if limits_reached(limits):
+                break
+
             name = self._extract_name_from_entry(entry)
             stat = self._extract_stat_from_entry(root, entry)
 
             if self._is_dir_entry(root, entry):
+                directory = RemoteDirectory(path=root / name, stats=stat)
+
                 if not topdown:
-                    yield from self._walk(root=root / name, topdown=topdown, filters=filters, limits=limits)
+                    yield from self._walk(root=directory, topdown=topdown, filters=filters, limits=limits)
 
-                path = RemoteDirectory(path=root / name, stats=stat)
-                if match_all_filters(path, filters):
-                    dirs.append(RemoteDirectory(path=name, stats=stat))
+                if match_all_filters(directory, filters) and not limits_stop_at(directory, limits):
+                    dirs.append(directory)
+            elif self._is_file_entry(root, entry):
+                file = RemoteFile(path=root / name, stats=stat)
 
-                    if limits_stop_at(path, limits):
-                        break
-            else:
-                path = RemoteFile(path=root / name, stats=stat)
+                if match_all_filters(file, filters) and not limits_stop_at(file, limits):
+                    files.append(file)
 
-                if match_all_filters(path, filters):
-                    files.append(RemoteFile(path=name, stats=stat))
-
-                    if limits_stop_at(path, limits):
-                        break
-
-        if topdown:
+        if topdown and not limits_reached(limits):
             for name in dirs:
                 yield from self._walk(root=root / name, topdown=topdown, filters=filters, limits=limits)
 
@@ -527,22 +536,22 @@ class FileConnection(BaseFileConnection, FrozenModel):
             stat = self._extract_stat_from_entry(root, entry)
 
             if self._is_dir_entry(root, entry):
-                path = RemoteDirectory(path=root / name, stats=stat)
-                log.debug("|%s| Directory to remove: %s", self.__class__.__name__, path_repr(path))
-                self._remove_dir_recursive(path)
-                log.debug("|%s| Successfully removed directory '%s'", self.__class__.__name__, path)
-            else:
-                path = RemoteFile(path=root / name, stats=stat)
-                log.debug("|%s| File to remove: %s", self.__class__.__name__, path_repr(path))
-                self._remove_file(path)
-                log.debug("|%s| Successfully removed file '%s'", self.__class__.__name__, path)
+                directory = RemoteDirectory(path=root / name, stats=stat)
+                log.debug("|%s| Directory to remove: %s", self.__class__.__name__, path_repr(directory))
+                self._remove_dir_recursive(directory)
+                log.debug("|%s| Successfully removed directory '%s'", self.__class__.__name__, directory)
+            elif self._is_file_entry(root, entry):
+                file = RemoteFile(path=root / name, stats=stat)
+                log.debug("|%s| File to remove: %s", self.__class__.__name__, path_repr(file))
+                self._remove_file(file)
+                log.debug("|%s| Successfully removed file '%s'", self.__class__.__name__, file)
 
         self._remove_dir(root)
 
     @abstractmethod
-    def _scan_entries(self, path: RemotePath) -> list:
+    def _scan_entries(self, path: RemotePath) -> Iterable:
         """
-        The method returns a list that contains entries.
+        The method returns an iterator that returns entries.
 
         Entry is an object containing information about a path, like file or nested directory.
 
@@ -556,19 +565,19 @@ class FileConnection(BaseFileConnection, FrozenModel):
 
         Returns
         -------
-        list
-            List of the entries
+        Iterable
+            Iterable entries
 
         Examples
         --------
 
-        Get a list of entries:
+        Get a entries:
 
         .. code:: python
 
-            entry_list = connection._scan_entries(path="/a/path/to/the/directory")
+            entry_iterable = connection._scan_entries(path="/a/path/to/the/directory")
 
-            print(entry_list)
+            print(entry_iterable)
 
             [
                 {
@@ -602,19 +611,20 @@ class FileConnection(BaseFileConnection, FrozenModel):
 
         Get an entry name:
 
-        .. code:: python
-
-            for entry in connection._scan_entries(path="/a/path/to/the/directory"):
-                assert entry == {
-                    "created": "2023-12-08T18:33:39Z",
-                    "owner": None,
-                    "size": "23",
-                    "modified": "2023-12-08 18:33:20",
-                    "isdir": False,
-                    "path": "/path/to/the/file.txt",
-                }
-
-                assert entry._extract_name_from_entry(entry) == "file.txt"
+        >>> for entry in connection._scan_entries(path="/a/path/to/the/directory"):
+        ...     break
+        >>> entry
+        # arbitrary object returned from underlying client
+        {
+            'created': '2023-12-08T18:33:39Z',
+            'owner': None,
+            'size': 23,
+            'modified': '2023-12-08 18:33:20',
+            'isdir': False,
+            'path': '/path/to/the/directory/file.txt',
+        }
+        >>> entry._extract_name_from_entry(entry)
+        'file.txt'
         """
 
     @abstractmethod
@@ -642,10 +652,19 @@ class FileConnection(BaseFileConnection, FrozenModel):
 
         Show if the entry is a directory:
 
-        .. code:: python
-
-            for component in connection._scan_entries(path="/a/path/to/the/directory"):
-                assert connection._is_dir_entry(root="/a/path/to/the/directory", entry) == True
+        >>> for entry in connection._scan_entries(path="/a/path/to/the/directory"):
+        ...     break
+        >>> entry
+        {
+            'created': '2023-12-08T18:33:39Z',
+            'owner': None,
+            'size': 0,
+            'modified': '2023-12-08 18:33:20',
+            'isdir': True,
+            'path': '/path/to/the/directory',
+        }
+        >>> connection._is_dir_entry(root="/a/path/to/the/directory", entry)
+        True
         """
 
     @abstractmethod
@@ -673,10 +692,19 @@ class FileConnection(BaseFileConnection, FrozenModel):
 
         Show if the entry is a file:
 
-        .. code:: python
-
-            for entry in connection._scan_entries(path="/a/path/to/the/directory"):
-                assert connection._is_file_entry(root="/a/path/to/the/directory", entry) == True
+        >>> for entry in connection._scan_entries(path="/a/path/to/the/directory"):
+        ...     break
+        >>> entry
+        {
+            'created': '2023-12-08T18:33:39Z',
+            'owner': None,
+            'size': 23,
+            'modified': '2023-12-08 18:33:20',
+            'isdir': False,
+            'path': '/path/to/the/directory/file.txt',
+        }
+        >>> connection._is_file_entry(root="/a/path/to/the/directory", entry)
+        True
         """
 
     @abstractmethod
@@ -707,15 +735,25 @@ class FileConnection(BaseFileConnection, FrozenModel):
 
         Get statistics object from the entry:
 
-        .. code:: python
-
-            for entry in connection._scan_entries(path="/a/path/to/the/directory"):
-                stat = connection._extract_stat_from_entry(root="/a/path/to/the/directory", entry)
-
-                assert stat == RemotePathStat(
-                    st_size=23, st_mtime=1670517693.0, st_mode=None, st_uid=None, st_gid=None
-                )
-
+        >>> for entry in connection._scan_entries(path="/a/path/to/the/directory"):
+        ...     break
+        >>> entry
+        {
+            'created': '2023-12-08T18:33:39Z',
+            'owner': None,
+            'size': 23,
+            'modified': '2023-12-08 18:33:20',
+            'isdir': False,
+            'path': '/path/to/the/directory/file.txt',
+        }
+        >>> connection._extract_stat_from_entry(root="/a/path/to/the/directory", entry)
+        RemotePathStat(
+            st_size=23,
+            st_mtime=1670517693.0,
+            st_mode=None,
+            st_uid=None,
+            st_gid=None,
+        )
         """
 
     def _log_parameters(self):
