@@ -1,28 +1,29 @@
-#  Copyright 2023 MTS (Mobile Telesystems)
-#
-#  Licensed under the Apache License, Version 2.0 (the "License");
-#  you may not use this file except in compliance with the License.
-#  You may obtain a copy of the License at
-#
-#      http://www.apache.org/licenses/LICENSE-2.0
-#
-#  Unless required by applicable law or agreed to in writing, software
-#  distributed under the License is distributed on an "AS IS" BASIS,
-#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-#  See the License for the specific language governing permissions and
-#  limitations under the License.
-
+# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
+# SPDX-License-Identifier: Apache-2.0
 from __future__ import annotations
 
 import os
 import stat
 import textwrap
+from contextlib import suppress
 from logging import getLogger
 from typing import TYPE_CHECKING, Optional, Tuple
 
 from etl_entities.instance import Cluster, Host
-from pydantic import Field, FilePath, SecretStr, root_validator, validator
 
+try:
+    from pydantic.v1 import (
+        Field,
+        FilePath,
+        PrivateAttr,
+        SecretStr,
+        root_validator,
+        validator,
+    )
+except (ImportError, AttributeError):
+    from pydantic import Field, FilePath, SecretStr, PrivateAttr, root_validator, validator  # type: ignore[no-redef, assignment]
+
+from onetl._util.alias import avoid_alias
 from onetl.base import PathStatProtocol
 from onetl.connection.file_connection.file_connection import FileConnection
 from onetl.connection.file_connection.hdfs.slots import HDFSSlots
@@ -43,10 +44,10 @@ except (ImportError, NameError) as err:
             Cannot import module "hdfs".
 
             Since onETL v0.7.0 you should install package as follows:
-                pip install onetl[hdfs]
+                pip install "onetl[hdfs]"
 
             or
-                pip install onetl[files]
+                pip install "onetl[files]"
             """,
         ).strip(),
     ) from err
@@ -67,19 +68,19 @@ class HDFS(FileConnection, RenameDirMixin):
 
         .. code:: bash
 
-            pip install onetl[hdfs]
+            pip install "onetl[hdfs]"
 
             # or
-            pip install onetl[files]
+            pip install "onetl[files]"
 
-        See :ref:`files-install` instruction for more details.
+        See :ref:`install-files` installation instruction for more details.
 
     .. note::
 
         To access Hadoop cluster with Kerberos installed, you should have ``kinit`` executable
         in some path in ``PATH`` environment variable.
 
-        See onETL :ref:`kerberos-install` instruction for more details.
+        See :ref:`install-kerberos` instruction for more details.
 
     Parameters
     ----------
@@ -96,6 +97,8 @@ class HDFS(FileConnection, RenameDirMixin):
 
             You should pass at least one of these arguments: ``cluster``, ``host``.
 
+        .. versionadded:: 0.7.0
+
     host : str, optional
         Hadoop namenode host. For example: ``namenode1.domain.com``.
 
@@ -148,67 +151,52 @@ class HDFS(FileConnection, RenameDirMixin):
     Examples
     --------
 
-    HDFS connection with password:
+    .. tabs::
 
-    .. code:: python
+        .. code-tab:: py Create HDFS connection with user+password
 
-        from onetl.connection import HDFS
+            from onetl.connection import HDFS
 
-        hdfs = HDFS(
-            host="namenode1.domain.com",
-            user="someuser",
-            password="*****",
-        ).check()
+            hdfs = HDFS(
+                host="namenode1.domain.com",
+                user="someuser",
+                password="*****",
+            ).check()
 
-    HDFS connection with keytab:
+        .. code-tab:: py Create HDFS connection with user+keytab
 
-    .. code:: python
+            from onetl.connection import HDFS
 
-        from onetl.connection import HDFS
+            hdfs = HDFS(
+                host="namenode1.domain.com",
+                user="someuser",
+                keytab="/path/to/keytab",
+            ).check()
 
-        hdfs = HDFS(
-            host="namenode1.domain.com",
-            user="someuser",
-            keytab="/path/to/keytab",
-        ).check()
+        .. code-tab:: py Create HDFS connection without auth
 
-    HDFS file connection initialization without auth (HDFS without Kerberos support)
+            from onetl.connection import HDFS
 
-    .. code:: python
+            hdfs = HDFS(host="namenode1.domain.com").check()
 
-        from onetl.connection import HDFS
+        .. tab:: Use cluster name to detect active namenode
 
-        hdfs = HDFS(host="namenode1.domain.com").check()
+            Can be used only if some third-party plugin provides :ref:`hdfs-slots` implementation
 
-    HDFS connection with both cluster and host names:
+            .. code:: python
 
-    .. code:: python
+                from onetl.connection import HDFS
 
-        from onetl.connection import HDFS
-
-        hdfs = HDFS(
-            cluster="rnd-dwh",
-            host="namenode1.domain.com",
-            user="someuser",
-            password="*****",
-        ).check()
-
-    HDFS connection with cluster name only:
-
-    .. code:: python
-
-        from onetl.connection import HDFS
-
-        hdfs = HDFS(
-            cluster="rnd-dwh",
-            user="someuser",
-            password="*****",
-        ).check()
+                hdfs = HDFS(
+                    cluster="rnd-dwh",
+                    user="someuser",
+                    password="*****",
+                ).check()
     """
 
     cluster: Optional[Cluster] = None
     host: Optional[Host] = None
-    webhdfs_port: int = Field(alias="port", default=50070)
+    webhdfs_port: int = Field(alias=avoid_alias("port"), default=50070)  # type: ignore[literal-required]
     user: Optional[str] = None
     password: Optional[SecretStr] = None
     keytab: Optional[FilePath] = None
@@ -218,6 +206,8 @@ class HDFS(FileConnection, RenameDirMixin):
     # TODO: remove in v1.0.0
     slots = Slots
 
+    _active_host: Optional[Host] = PrivateAttr(default=None)
+
     @slot
     @classmethod
     def get_current(cls, **kwargs):
@@ -231,6 +221,8 @@ class HDFS(FileConnection, RenameDirMixin):
             Can be used only if there are a some hooks bound to slot
             :obj:`Slots.get_current_cluster <onetl.connection.file_connection.hdfs.slots.HDFSSlots.get_current_cluster>`
 
+        .. versionadded:: 0.7.0
+
         Parameters
         ----------
         user : str
@@ -268,10 +260,23 @@ class HDFS(FileConnection, RenameDirMixin):
             return self.cluster
         return f"hdfs://{self.host}:{self.webhdfs_port}"
 
+    def __str__(self):
+        if self.cluster:
+            return f"{self.__class__.__name__}[{self.cluster}]"
+        return f"{self.__class__.__name__}[{self.host}:{self.webhdfs_port}]"
+
     @slot
     def path_exists(self, path: os.PathLike | str) -> bool:
         return self.client.status(os.fspath(path), strict=False)
 
+    @slot
+    def close(self):
+        super().close()
+
+        with suppress(Exception):
+            self._active_host = None
+        return self
+
     @validator("user", pre=True)
     def _validate_packages(cls, user):
         if user:
@@ -284,10 +289,7 @@ class HDFS(FileConnection, RenameDirMixin):
                         Cannot import module "hdfs.ext.kerberos".
 
                         Since onETL v0.7.0 you should install package as follows:
-                            pip install onetl[hdfs,kerberos]
-
-                        or
-                            pip install onetl[all]
+                            pip install "onetl[hdfs,kerberos]"
 
                         You should also have Kerberos libraries installed to OS,
                         specifically ``kinit`` executable.
@@ -377,7 +379,7 @@ class HDFS(FileConnection, RenameDirMixin):
             raise RuntimeError(f"Cannot get list of namenodes for a cluster {self.cluster!r}")
 
         nodes_len = len(namenodes)
-        for i, namenode in enumerate(namenodes):
+        for i, namenode in enumerate(namenodes, start=1):
             log.debug("|%s|   Trying namenode %r (%d of %d) ...", class_name, namenode, i, nodes_len)
             if self.Slots.is_namenode_active(namenode, self.cluster):
                 log.info("|%s|     Node %r is active!", class_name, namenode)
@@ -411,10 +413,13 @@ class HDFS(FileConnection, RenameDirMixin):
 
         raise RuntimeError(f"Host {self.host!r} is not an active namenode")
 
-    def _get_client(self) -> Client:
-        host = self._get_host()
-        conn_str = f"http://{host}:{self.webhdfs_port}"  # NOSONAR
+    def _get_conn_str(self) -> str:
+        # cache active host to reduce number of requests.
+        if not self._active_host:
+            self._active_host = self._get_host()
+        return f"http://{self._active_host}:{self.webhdfs_port}"
 
+    def _get_client(self) -> Client:
         if self.user and (self.keytab or self.password):
             from hdfs.ext.kerberos import KerberosClient  # noqa: F811
 
@@ -423,10 +428,13 @@ class HDFS(FileConnection, RenameDirMixin):
                 keytab=self.keytab,
                 password=self.password.get_secret_value() if self.password else None,
             )
+            # checking if namenode is active requires a Kerberos ticket
+            conn_str = self._get_conn_str()
             client = KerberosClient(conn_str, timeout=self.timeout)
         else:
             from hdfs import InsecureClient  # noqa: F401, WPS442, F811
 
+            conn_str = self._get_conn_str()
             client = InsecureClient(conn_str, user=self.user)
 
         return client
@@ -541,7 +549,7 @@ class HDFS(FileConnection, RenameDirMixin):
             st_mtime=entry_stat["modificationTime"] / 1000,  # HDFS uses timestamps with milliseconds
             st_uid=entry_stat["owner"],
             st_gid=entry_stat["group"],
-            st_mode=int(entry_stat["permission"], 8) | stat.S_IFDIR
-            if entry_stat["type"] == "DIRECTORY"
-            else stat.S_IFREG,
+            st_mode=(
+                int(entry_stat["permission"], 8) | stat.S_IFDIR if entry_stat["type"] == "DIRECTORY" else stat.S_IFREG
+            ),
         )
