@@ -1,32 +1,25 @@
-#  Copyright 2023 MTS (Mobile Telesystems)
-#
-#  Licensed under the Apache License, Version 2.0 (the "License");
-#  you may not use this file except in compliance with the License.
-#  You may obtain a copy of the License at
-#
-#      http://www.apache.org/licenses/LICENSE-2.0
-#
-#  Unless required by applicable law or agreed to in writing, software
-#  distributed under the License is distributed on an "AS IS" BASIS,
-#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-#  See the License for the specific language governing permissions and
-#  limitations under the License.
-
+# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
+# SPDX-License-Identifier: Apache-2.0
 from __future__ import annotations
 
 import logging
 import os
+from contextlib import suppress
 from typing import TYPE_CHECKING, ClassVar, List, Optional
 
 from etl_entities.instance import Host
-from pydantic import SecretStr, root_validator, validator
+
+try:
+    from pydantic.v1 import SecretStr, root_validator, validator
+except (ImportError, AttributeError):
+    from pydantic import SecretStr, root_validator, validator  # type: ignore[no-redef, assignment]
+
 from typing_extensions import Literal
 
-from onetl._internal import stringify
-from onetl._util.hadoop import get_hadoop_config, get_hadoop_version
+from onetl._util.hadoop import get_hadoop_config
 from onetl._util.java import try_import_java_class
 from onetl._util.scala import get_default_scala_version
-from onetl._util.spark import get_spark_version
+from onetl._util.spark import get_client_info, get_spark_version, stringify
 from onetl._util.version import Version
 from onetl.base import (
     BaseReadableFileFormat,
@@ -59,30 +52,9 @@ class SparkS3(SparkFileDFConnection):
     Based on `Hadoop-AWS module <https://hadoop.apache.org/docs/current3/hadoop-aws/tools/hadoop-aws/index.html>`_
     and `Spark integration with Cloud Infrastructures <https://spark.apache.org/docs/latest/cloud-integration.html>`_.
 
-    .. dropdown:: Version compatibility
+    .. seealso::
 
-        * Spark versions: 3.2.x - 3.4.x (only with Hadoop 3.x libraries)
-        * Java versions: 8 - 20
-
-    .. warning::
-
-        See :ref:`spark-s3-troubleshooting` guide.
-
-    .. warning::
-
-        To use SparkS3 connector you should have PySpark installed (or injected to ``sys.path``)
-        BEFORE creating the connector instance.
-
-        You can install PySpark as follows:
-
-        .. code:: bash
-
-            pip install onetl[spark]  # latest PySpark version
-
-            # or
-            pip install onetl pyspark=3.4.1  # pass specific PySpark version
-
-        See :ref:`spark-install` instruction for more details.
+        Before using this connector please take into account :ref:`spark-s3-prerequisites`
 
     .. note::
 
@@ -91,6 +63,8 @@ class SparkS3(SparkFileDFConnection):
         Does NOT support file operations, like create, delete, rename, etc. For these operations,
         use :obj:`S3 <onetl.connection.file_connection.s3.S3>` connection.
 
+    .. versionadded:: 0.9.0
+
     Parameters
     ----------
     host : str
@@ -149,66 +123,65 @@ class SparkS3(SparkFileDFConnection):
     Examples
     --------
 
-    Create connection to S3 to work with dataframes.
+    .. tabs::
 
-    Connect to bucket as subdomain (``my-bucket.domain.com``):
+        .. code-tab:: py Create S3 connection with bucket as subdomain (``my-bucket.domain.com``):
 
-    .. code:: python
+            from onetl.connection import SparkS3
+            from pyspark.sql import SparkSession
 
-        from onetl.connection import SparkS3
-        from pyspark.sql import SparkSession
-
-        # Create Spark session with Hadoop AWS libraries loaded
-        maven_packages = SparkS3.get_packages(spark_version="3.4.1")
-        spark = (
-            SparkSession.builder.appName("spark-app-name")
-            .config("spark.jars.packages", ",".join(maven_packages))
-            .config("spark.hadoop.fs.s3a.committer.magic.enabled", "true")
-            .config("spark.hadoop.fs.s3a.committer.name", "magic")
-            .config(
-                "spark.hadoop.mapreduce.outputcommitter.factory.scheme.s3a",
-                "org.apache.hadoop.fs.s3a.commit.S3ACommitterFactory",
+            # Create Spark session with Hadoop AWS libraries loaded
+            maven_packages = SparkS3.get_packages(spark_version="3.5.6")
+            # Some packages are not used, but downloading takes a lot of time. Skipping them.
+            excluded_packages = SparkS3.get_exclude_packages()
+            spark = (
+                SparkSession.builder.appName("spark-app-name")
+                .config("spark.jars.packages", ",".join(maven_packages))
+                .config("spark.jars.excludes", ",".join(excluded_packages))
+                .config("spark.hadoop.fs.s3a.committer.magic.enabled", "true")
+                .config("spark.hadoop.fs.s3a.committer.name", "magic")
+                .config(
+                    "spark.hadoop.mapreduce.outputcommitter.factory.scheme.s3a",
+                    "org.apache.hadoop.fs.s3a.commit.S3ACommitterFactory",
+                )
+                .config(
+                    "spark.sql.parquet.output.committer.class",
+                    "org.apache.spark.internal.io.cloud.BindingParquetOutputCommitter",
+                )
+                .config(
+                    "spark.sql.sources.commitProtocolClass",
+                    "org.apache.spark.internal.io.cloud.PathOutputCommitProtocol",
+                )
+                .getOrCreate()
             )
-            .config(
-                "spark.sql.parquet.output.committer.class",
-                "org.apache.spark.internal.io.cloud.BindingParquetOutputCommitter",
-            )
-            .config(
-                "spark.sql.sources.commitProtocolClass",
-                "org.apache.spark.internal.io.cloud.PathOutputCommitProtocol",
-            )
-            .getOrCreate()
-        )
 
-        # Create connection
-        s3 = SparkS3(
-            host="domain.com",
-            protocol="http",
-            bucket="my-bucket",
-            access_key="ACCESS_KEY",
-            secret_key="SECRET_KEY",
-            spark=spark,
-        ).check()
+            # Create connection
+            s3 = SparkS3(
+                host="domain.com",
+                protocol="http",
+                bucket="my-bucket",
+                access_key="ACCESS_KEY",
+                secret_key="SECRET_KEY",
+                spark=spark,
+            ).check()
 
-    Connect to bucket as subpath (``domain.com/my-bucket``):
+        .. code-tab:: py Create S3 connection with bucket as subpath (``domain.com/my-bucket``)
 
-    .. code:: python
+            # Create Spark session with Hadoop AWS libraries loaded
+            ...
 
-        # Create Spark session with Hadoop AWS libraries loaded
-        ...
-
-        # Create connection
-        s3 = SparkS3(
-            host="domain.com",
-            protocol="http",
-            bucket="my-bucket",
-            access_key="ACCESS_KEY",
-            secret_key="SECRET_KEY",
-            extra={
-                "path.style.access": True,
-            },
-            spark=spark,
-        ).check()
+            # Create connection
+            s3 = SparkS3(
+                host="domain.com",
+                protocol="http",
+                bucket="my-bucket",
+                access_key="ACCESS_KEY",
+                secret_key="SECRET_KEY",
+                extra={
+                    "path.style.access": True,  # <---
+                },
+                spark=spark,
+            ).check()
     """
 
     Extra = SparkS3Extra
@@ -223,7 +196,11 @@ class SparkS3(SparkFileDFConnection):
     region: Optional[str] = None
     extra: SparkS3Extra = SparkS3Extra()
 
-    _ROOT_CONFIG_KEYS: ClassVar[List[str]] = ["committer.magic.enabled", "committer.name"]
+    _ROOT_CONFIG_KEYS: ClassVar[List[str]] = [
+        "committer.magic.enabled",
+        "committer.name",
+        "user.agent.prefix",
+    ]
 
     @slot
     @classmethod
@@ -235,6 +212,8 @@ class SparkS3(SparkFileDFConnection):
         """
         Get package names to be downloaded by Spark. |support_hooks|
 
+        .. versionadded:: 0.9.0
+
         Parameters
         ----------
         spark_version : str
@@ -252,18 +231,59 @@ class SparkS3(SparkFileDFConnection):
 
             from onetl.connection import SparkS3
 
-            SparkS3.get_packages(spark_version="3.4.1")
-            SparkS3.get_packages(spark_version="3.4.1", scala_version="2.12")
+            SparkS3.get_packages(spark_version="3.5.6")
+            SparkS3.get_packages(spark_version="3.5.6", scala_version="2.12")
 
         """
 
-        spark_ver = Version.parse(spark_version)
-        if spark_ver.major < 3:
-            # https://issues.apache.org/jira/browse/SPARK-23977
-            raise ValueError(f"Spark version must be at least 3.x, got {spark_ver}")
+        spark_ver = Version(spark_version).min_digits(3)
+        scala_ver = Version(scala_version).min_digits(2) if scala_version else get_default_scala_version(spark_ver)
+        # https://mvnrepository.com/artifact/org.apache.spark/spark-hadoop-cloud
+        return [f"org.apache.spark:spark-hadoop-cloud_{scala_ver.format('{0}.{1}')}:{spark_ver.format('{0}.{1}.{2}')}"]
 
-        scala_ver = Version.parse(scala_version) if scala_version else get_default_scala_version(spark_ver)
-        return [f"org.apache.spark:spark-hadoop-cloud_{scala_ver.digits(2)}:{spark_ver.digits(3)}"]
+    @slot
+    @classmethod
+    def get_exclude_packages(cls) -> list[str]:
+        """
+        Get package names to be excluded by Spark. |support_hooks|
+
+        .. versionadded:: 0.13.0
+
+        Examples
+        --------
+
+        .. code:: python
+
+            from onetl.connection import SparkS3
+
+            SparkS3.get_exclude_packages()
+
+        """
+
+        return [
+            # heavy and not used
+            "com.google.cloud.bigdataoss:gcs-connector",
+            "org.apache.hadoop:hadoop-aliyun",
+            "org.apache.hadoop:hadoop-azure-datalake",
+            "org.apache.hadoop:hadoop-azure",
+            "org.apache.hadoop:hadoop-cos",
+            "org.apache.hadoop:hadoop-openstack",
+            "org.apache.hadoop:hadoop-huaweicloud",
+            # already a part of Spark bundle
+            "org.apache.hadoop:hadoop-client-api",
+            "org.apache.hadoop:hadoop-client-runtime",
+            "com.fasterxml.jackson.core:jackson-annotations",
+            "com.fasterxml.jackson.core:jackson-core",
+            "com.fasterxml.jackson.core:jackson-databind",
+            "com.google.code.findbugs:jsr305",
+            "commons-codec:commons-codec",
+            "commons-logging:commons-logging",
+            "joda-time:joda-time",
+            "org.apache.httpcomponents:httpclient",
+            "org.apache.httpcomponents:httpcore",
+            "org.slf4j:slf4j-api",
+            "org.xerial.snappy:snappy-java",
+        ]
 
     @slot
     def path_from_string(self, path: os.PathLike | str) -> RemotePath:
@@ -271,7 +291,10 @@ class SparkS3(SparkFileDFConnection):
 
     @property
     def instance_url(self):
-        return f"s3://{self.host}:{self.port}"
+        return f"s3://{self.host}:{self.port}/{self.bucket}"
+
+    def __str__(self):
+        return f"S3[{self.host}:{self.port}/{self.bucket}]"
 
     def __enter__(self):
         return self
@@ -311,9 +334,13 @@ class SparkS3(SparkFileDFConnection):
             connection.close()
 
         """
-        self._reset_hadoop_conf()
+        with suppress(Exception):
+            self._reset_hadoop_conf()
         return self
 
+    # Do not all __del__ with calling .close(), like other connections,
+    # because this can influence dataframes created by this connection
+
     @slot
     def check(self):
         self._patch_hadoop_conf()
@@ -357,25 +384,16 @@ class SparkS3(SparkFileDFConnection):
         try:
             try_import_java_class(spark, java_class)
         except Exception as e:
-            spark_version = get_spark_version(spark).digits(3)
+            spark_version = get_spark_version(spark).format("{0}.{1}.{2}")
             msg = MISSING_JVM_CLASS_MSG.format(
                 java_class=java_class,
                 package_source=cls.__name__,
                 args=f"spark_version='{spark_version}'",
             )
-            if log.isEnabledFor(logging.DEBUG):
-                log.debug("Missing Java class", exc_info=e, stack_info=True)
             raise ValueError(msg) from e
 
         return spark
 
-    @validator("spark")
-    def _validate_hadoop_version(cls, spark: SparkSession) -> SparkSession:
-        hadoop_version = get_hadoop_version(spark)
-        if hadoop_version.major < 3:
-            raise ValueError(f"Only Hadoop 3.x libraries are supported, got {hadoop_version}")
-        return spark
-
     def _get_hadoop_config_prefix(self) -> str:
         return f"fs.s3a.bucket.{self.bucket}"
 
@@ -411,6 +429,14 @@ class SparkS3(SparkFileDFConnection):
         for key, value in self.extra.dict(by_alias=True, exclude_none=True).items():
             conf[f"{prefix}.{key}"] = value
 
+        # https://hadoop.apache.org/docs/r3.4.1/hadoop-aws/tools/hadoop-aws/index.html
+        user_agent = conf.get(f"{prefix}.user.agent.prefix")
+        client_info = get_client_info(self.spark)
+        if user_agent:
+            conf[f"{prefix}.user.agent.prefix"] = f"{user_agent} {client_info}"
+        else:
+            conf[f"{prefix}.user.agent.prefix"] = client_info
+
         self._fix_root_conf(conf, prefix)
         return stringify(conf)
 
