@@ -1,17 +1,5 @@
-#  Copyright 2023 MTS (Mobile Telesystems)
-#
-#  Licensed under the Apache License, Version 2.0 (the "License");
-#  you may not use this file except in compliance with the License.
-#  You may obtain a copy of the License at
-#
-#      http://www.apache.org/licenses/LICENSE-2.0
-#
-#  Unless required by applicable law or agreed to in writing, software
-#  distributed under the License is distributed on an "AS IS" BASIS,
-#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-#  See the License for the specific language governing permissions and
-#  limitations under the License.
-
+# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
+# SPDX-License-Identifier: Apache-2.0
 from __future__ import annotations
 
 import logging
@@ -21,7 +9,11 @@ from enum import Enum
 from typing import Iterable, List, Optional, Tuple
 
 from ordered_set import OrderedSet
-from pydantic import Field, validator
+
+try:
+    from pydantic.v1 import Field, PrivateAttr, validator
+except (ImportError, AttributeError):
+    from pydantic import Field, PrivateAttr, validator  # type: ignore[no-redef, assignment]
 
 from onetl.base import BaseFileConnection, BaseFileFilter, BaseFileLimit
 from onetl.base.path_protocol import PathProtocol, PathWithStatsProtocol
@@ -75,6 +67,8 @@ class FileMover(FrozenModel):
 
         This class does **not** support read strategies.
 
+    .. versionadded:: 0.8.0
+
     Parameters
     ----------
     connection : :obj:`onetl.connection.FileConnection`
@@ -97,61 +91,62 @@ class FileMover(FrozenModel):
         See :ref:`file-limits`
 
     options : :obj:`~FileMover.Options`  | dict | None, default: ``None``
-        File moving options. See :obj:`~FileMover.Options`
+        File moving options. See :obj:`FileMover.Options <onetl.file.file_mover.options.FileMoverOptions>`
 
     Examples
     --------
-    Simple Mover creation
 
-    .. code:: python
+    .. tabs::
 
-        from onetl.connection import SFTP
-        from onetl.file import FileMover
+        .. code-tab:: py Minimal example
 
-        sftp = SFTP(...)
+            from onetl.connection import SFTP
+            from onetl.file import FileMover
 
-        # create mover
-        mover = FileMover(
-            connection=sftp,
-            source_path="/path/to/source/dir",
-            target_path="/path/to/target/dir",
-        )
+            sftp = SFTP(...)
 
-        # move files from "/path/to/source/dir" to "/path/to/target/dir"
-        mover.run()
+            # create mover
+            mover = FileMover(
+                connection=sftp,
+                source_path="/path/to/source/dir",
+                target_path="/path/to/target/dir",
+            )
 
-    Mover with all parameters
+            # move files from "/path/to/source/dir" to "/path/to/target/dir"
+            mover.run()
 
-    .. code:: python
+        .. code-tab:: py Full example
 
-        from onetl.connection import SFTP
-        from onetl.file import FileMover
-        from onetl.file.filter import Glob, ExcludeDir
-        from onetl.file.limit import MaxFilesCount
+            from onetl.connection import SFTP
+            from onetl.file import FileMover
+            from onetl.file.filter import Glob, ExcludeDir
+            from onetl.file.limit import MaxFilesCount, TotalFilesSize
 
-        sftp = SFTP(...)
+            sftp = SFTP(...)
 
-        # create mover with a bunch of options
-        mover = FileMover(
-            connection=sftp,
-            source_path="/path/to/source/dir",
-            target_path="/path/to/target/dir",
-            filters=[
-                Glob("*.txt"),
-                ExcludeDir("/path/to/source/dir/exclude"),
-            ],
-            limits=[MaxFilesCount(100)],
-            options=FileMover.Options(if_exists="replace_file"),
-        )
+            # create mover with a bunch of options
+            mover = FileMover(
+                connection=sftp,
+                source_path="/path/to/source/dir",
+                target_path="/path/to/target/dir",
+                filters=[
+                    Glob("*.txt"),
+                    ExcludeDir("/path/to/source/dir/exclude"),
+                ],
+                limits=[MaxFilesCount(100), TotalFileSize("10GiB")],
+                options=FileMover.Options(if_exists="replace_file"),
+            )
 
-        # move files from "/path/to/source/dir" to "/path/to/target/dir",
-        # but only *.txt files
-        # excluding files from "/path/to/source/dir/exclude" directory
-        # and stop before downloading 101 file
-        mover.run()
+            # move files from "/path/to/source/dir" to "/path/to/target/dir",
+            # but only *.txt files
+            # excluding files from "/path/to/source/dir/exclude" directory
+            # and stop before downloading 101 file
+            mover.run()
 
     """
 
+    Options = FileMoverOptions
+
     connection: BaseFileConnection
 
     target_path: RemotePath
@@ -162,13 +157,15 @@ class FileMover(FrozenModel):
 
     options: FileMoverOptions = FileMoverOptions()
 
-    Options = FileMoverOptions
+    _connection_checked: bool = PrivateAttr(default=False)
 
     @slot
     def run(self, files: Iterable[str | os.PathLike] | None = None) -> MoveResult:  # noqa: WPS231
         """
         Method for moving files from source to target directory. |support_hooks|
 
+        .. versionadded:: 0.8.0
+
         Parameters
         ----------
 
@@ -183,12 +180,12 @@ class FileMover(FrozenModel):
 
         Returns
         -------
-        moved_files : :obj:`MoveResult <onetl.file.file_mover.move_result.MoveResult>`
+        :obj:`MoveResult <onetl.file.file_mover.move_result.MoveResult>`
 
             Move result object
 
         Raises
-        -------
+        ------
         :obj:`onetl.exception.DirectoryNotFoundError`
 
             ``source_path`` does not found
@@ -200,94 +197,97 @@ class FileMover(FrozenModel):
         Examples
         --------
 
-        Move files from ``source_path``
+        Move files from ``source_path``:
 
-        .. code:: python
-
-            from onetl.impl import RemoteFile, RemotePath
-            from onetl.file import FileMover
-
-            mover = FileMover(source_path="/source", target_path="/target", ...)
-            moved_files = mover.run()
-
-            assert moved_files.successful == {
+        >>> from onetl.file import FileMover
+        >>> mover = FileMover(source_path="/source", target_path="/target", ...)
+        >>> move_result = mover.run()
+        >>> move_result
+        MoveResult(
+            successful=FileSet([
                 RemoteFile("/target/file1.txt"),
                 RemoteFile("/target/file2.txt"),
-                RemoteFile("/target/nested/path/file3.txt"),  # directory structure is preserved
-            }
-            assert moved_files.failed == {FailedRemoteFile("/source/failed.file")}
-            assert moved_files.skipped == {RemoteFile("/source/already.exists")}
-            assert moved_files.missing == {RemotePath("/source/missing.file")}
+                # directory structure is preserved
+                RemoteFile("/target/nested/path/file3.txt"),
+            ]),
+            failed=FileSet([
+                FailedRemoteFile("/source/failed.file"),
+            ]),
+            skipped=FileSet([
+                RemoteFile("/source/already.exists"),
+            ]),
+            missing=FileSet([
+                RemotePath("/source/missing.file"),
+            ]),
+        )
 
-        Move only certain files from ``source_path``
+        Move only certain files from ``source_path``:
 
-        .. code:: python
-
-            from onetl.impl import RemoteFile
-            from onetl.file import FileMover
-
-            mover = FileMover(source_path="/source", target_path="/target", ...)
-
-            # paths could be relative or absolute, but all should be in "/source"
-            moved_files = mover.run(
-                [
-                    "/source/file1.txt",
-                    "/source/nested/path/file3.txt",
-                    # excluding "/source/file2.txt"
-                ]
-            )
-
-            assert moved_files.successful == {
+        >>> from onetl.file import FileMover
+        >>> mover = FileMover(source_path="/source", target_path="/target", ...)
+        >>> # paths could be relative or absolute, but all should be in "/source"
+        >>> move_result = mover.run(
+        ...     [
+        ...         "/source/file1.txt",
+        ...         "/source/nested/path/file3.txt",
+        ...         # excluding "/source/file2.txt"
+        ...     ]
+        ... )
+        >>> move_result
+        MoveResult(
+            successful=FileSet([
                 RemoteFile("/target/file1.txt"),
-                RemoteFile("/target/nested/path/file3.txt"),  # directory structure is preserved
-            }
-            assert not moved_files.failed
-            assert not moved_files.skipped
-            assert not moved_files.missing
+                # directory structure is preserved
+                RemoteFile("/target/nested/path/file3.txt"),
+            ]),
+            failed=FileSet([]),
+            skipped=FileSet([]),
+            missing=FileSet([]),
+        )
 
-        Move certain files from any folder
+        Move certain files from any folder:
 
-        .. code:: python
-
-            from onetl.impl import RemoteFile
-            from onetl.file import FileMover
-
-            mover = FileMover(target_path="/target", ...)  # no source_path set
-
-            # only absolute paths
-            moved_files = mover.run(
-                [
-                    "/remote/file1.txt",
-                    "/any/nested/path/file3.txt",
-                ]
-            )
-
-            assert moved_files.successful == {
+        >>> from onetl.file import FileMover
+        >>> mover = FileMover(target_path="/target", ...)  # no source_path set
+        >>> # only absolute paths
+        >>> move_result = mover.run(
+        ...     [
+        ...         "/remote/file1.txt",
+        ...         "/any/nested/path/file3.txt",
+        ...     ]
+        ... )
+        >>> move_result
+        MoveResult(
+            successful=FileSet([
                 RemoteFile("/target/file1.txt"),
-                RemoteFile("/target/file3.txt"),
                 # directory structure is NOT preserved without source_path
-            }
-            assert not moved_files.failed
-            assert not moved_files.skipped
-            assert not moved_files.missing
+                RemoteFile("/target/file3.txt"),
+            ]),
+            failed=FileSet([]),
+            skipped=FileSet([]),
+            missing=FileSet([]),
+        )
         """
 
+        entity_boundary_log(log, f"{self.__class__.__name__}.run() starts")
+
         if files is None and not self.source_path:
             raise ValueError("Neither file list nor `source_path` are passed")
 
-        self._log_parameters(files)
+        if not self._connection_checked:
+            self._log_parameters(files)
 
-        # Check everything
-        self.connection.check()
-        self._check_target_path()
-        log_with_indent(log, "")
+            self.connection.check()
+            self._check_target_path()
+            log_with_indent(log, "")
 
-        if self.source_path:
-            self._check_source_path()
+            if self.source_path:
+                self._check_source_path()
+
+            self._connection_checked = True
 
         if files is None:
-            log.info("|%s| File list is not passed to `run` method", self.__class__.__name__)
-
+            log.debug("|%s| File list is not passed to `run` method", self.__class__.__name__)
             files = self.view_files()
 
         if not files:
@@ -303,6 +303,7 @@ class FileMover(FrozenModel):
 
         result = self._move_files(to_move)
         self._log_result(result)
+        entity_boundary_log(log, f"{self.__class__.__name__}.run() ends", char="-")
         return result
 
     @slot
@@ -311,8 +312,10 @@ class FileMover(FrozenModel):
         Get file list in the ``source_path``,
         after ``filter`` and ``limit`` applied (if any). |support_hooks|
 
+        .. versionadded:: 0.8.0
+
         Raises
-        -------
+        ------
         :obj:`onetl.exception.DirectoryNotFoundError`
 
             ``source_path`` does not found
@@ -329,33 +332,32 @@ class FileMover(FrozenModel):
         Examples
         --------
 
-        View files
+        View files:
 
-        .. code:: python
-
-            from onetl.impl import RemoteFile
-            from onetl.file import FileMover
-
-            mover = FileMover(source_path="/remote", ...)
-
-            view_files = mover.view_files()
-
-            assert view_files == {
-                RemoteFile("/remote/file1.txt"),
-                RemoteFile("/remote/file3.txt"),
-                RemoteFile("/remote/nested/path/file3.txt"),
-            }
+        >>> from onetl.file import FileMover
+        >>> mover = FileMover(source_path="/remote", ...)
+        >>> mover.view_files()
+        FileSet([
+            RemoteFile("/remote/file1.txt"),
+            RemoteFile("/remote/file2.txt"),
+            RemoteFile("/remote/nested/path/file3.txt"),
+        ])
         """
 
+        if not self.source_path:
+            raise ValueError("Cannot call `.view_files()` without `source_path`")
+
         log.debug("|%s| Getting files list from path '%s'", self.connection.__class__.__name__, self.source_path)
 
-        self._check_source_path()
+        if not self._connection_checked:
+            self._check_source_path()
+
         result = FileSet()
 
         try:
-            for root, _dirs, files in self.connection.walk(self.source_path, filters=self.filters, limits=self.limits):
+            for _root, _dirs, files in self.connection.walk(self.source_path, filters=self.filters, limits=self.limits):
                 for file in files:
-                    result.append(RemoteFile(path=root / file, stats=file.stats))
+                    result.append(file)
 
         except Exception as e:
             raise RuntimeError(
@@ -365,8 +367,6 @@ class FileMover(FrozenModel):
         return result
 
     def _log_parameters(self, files: Iterable[str | os.PathLike] | None = None) -> None:
-        entity_boundary_log(log, msg="FileMover starts")
-
         connection_class = self.connection.__class__.__name__
         log.info("|%s| -> |%s| Moving files using parameters:", connection_class, connection_class)
         log_with_indent(log, "source_path = %s", f"'{self.source_path}'" if self.source_path else "None")
@@ -566,4 +566,3 @@ class FileMover(FrozenModel):
         log_with_indent(log, "")
         log.info("|%s| Move result:", self.__class__.__name__)
         log_lines(log, str(result))
-        entity_boundary_log(log, msg=f"{self.__class__.__name__} ends", char="-")
