@@ -1,17 +1,5 @@
-#  Copyright 2023 MTS (Mobile Telesystems)
-#
-#  Licensed under the Apache License, Version 2.0 (the "License");
-#  you may not use this file except in compliance with the License.
-#  You may obtain a copy of the License at
-#
-#      http://www.apache.org/licenses/LICENSE-2.0
-#
-#  Unless required by applicable law or agreed to in writing, software
-#  distributed under the License is distributed on an "AS IS" BASIS,
-#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-#  See the License for the specific language governing permissions and
-#  limitations under the License.
-
+# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
+# SPDX-License-Identifier: Apache-2.0
 from __future__ import annotations
 
 import logging
@@ -21,9 +9,13 @@ from enum import Enum
 from typing import Iterable, Optional, Tuple
 
 from ordered_set import OrderedSet
-from pydantic import validator
 
-from onetl._internal import generate_temp_path
+try:
+    from pydantic.v1 import PrivateAttr, validator
+except (ImportError, AttributeError):
+    from pydantic import PrivateAttr, validator  # type: ignore[no-redef, assignment]
+
+from onetl._util.file import generate_temp_path
 from onetl.base import BaseFileConnection
 from onetl.base.path_protocol import PathWithStatsProtocol
 from onetl.base.pure_path_protocol import PurePathProtocol
@@ -71,6 +63,11 @@ class FileUploader(FrozenModel):
 
         This class does **not** support read strategies.
 
+    .. versionadded:: 0.1.0
+
+    .. versionchanged:: 0.8.0
+        Moved ``onetl.core.FileDownloader`` → ``onetl.file.FileDownloader``
+
     Parameters
     ----------
     connection : :obj:`onetl.connection.FileConnection`
@@ -85,6 +82,8 @@ class FileUploader(FrozenModel):
         Could be ``None``, but only if you pass absolute file paths directly to
         :obj:`~run` method
 
+        .. versionadded:: 0.3.0
+
     temp_path : os.PathLike or str, optional, default: ``None``
         If set, this path will be used for uploading a file, and then renaming it to the target file path.
         If ``None`` (default since v0.5.0) is passed, files are uploaded directly to ``target_path``.
@@ -103,44 +102,47 @@ class FileUploader(FrozenModel):
             Otherwise instead of ``rename``, remote OS will move file between filesystems,
             which is NOT atomic operation.
 
+        .. versionchanged:: 0.5.0
+            Default changed from ``/tmp`` to ``None``
+
     options : :obj:`~FileUploader.Options` | dict | None, default: ``None``
-        File upload options. See :obj:`~FileUploader.Options`
+        File upload options. See :obj:`FileUploader.Options <onetl.file.file_uploader.options.FileUploaderOptions>`
 
     Examples
     --------
-    Simple Uploader creation
 
-    .. code:: python
+    .. tabs::
 
-        from onetl.connection import HDFS
-        from onetl.file import FileUploader
+        .. code-tab:: py Minimal example
 
-        hdfs = HDFS(...)
+            from onetl.connection import HDFS
+            from onetl.file import FileUploader
 
-        uploader = FileUploader(
-            connection=hdfs,
-            target_path="/path/to/remote/source",
-        )
+            hdfs = HDFS(...)
 
-    Uploader with all parameters
+            uploader = FileUploader(
+                connection=hdfs,
+                target_path="/path/to/remote/source",
+            )
 
-    .. code:: python
+        .. code-tab:: py Full example
 
-        from onetl.connection import HDFS
-        from onetl.file import FileUploader
+            from onetl.connection import HDFS
+            from onetl.file import FileUploader
 
-        hdfs = HDFS(...)
-
-        uploader = FileUploader(
-            connection=hdfs,
-            target_path="/path/to/remote/source",
-            temp_path="/user/onetl",
-            local_path="/some/local/directory",
-            options=FileUploader.Options(delete_local=True, if_exists="overwrite"),
-        )
+            hdfs = HDFS(...)
 
+            uploader = FileUploader(
+                connection=hdfs,
+                target_path="/path/to/remote/source",
+                temp_path="/user/onetl",
+                local_path="/some/local/directory",
+                options=FileUploader.Options(delete_local=True, if_exists="overwrite"),
+            )
     """
 
+    Options = FileUploaderOptions
+
     connection: BaseFileConnection
 
     target_path: RemotePath
@@ -150,13 +152,15 @@ class FileUploader(FrozenModel):
 
     options: FileUploaderOptions = FileUploaderOptions()
 
-    Options = FileUploaderOptions
+    _connection_checked: bool = PrivateAttr(default=False)
 
     @slot
     def run(self, files: Iterable[str | os.PathLike] | None = None) -> UploadResult:
         """
         Method for uploading files to remote host. |support_hooks|
 
+        .. versionadded:: 0.1.0
+
         Parameters
         ----------
 
@@ -167,12 +171,12 @@ class FileUploader(FrozenModel):
 
         Returns
         -------
-        uploaded_files : :obj:`UploadResult <onetl.file.file_uploader.upload_result.UploadResult>`
+        :obj:`UploadResult <onetl.file.file_uploader.upload_result.UploadResult>`
 
             Upload result object
 
         Raises
-        -------
+        ------
         :obj:`onetl.exception.DirectoryNotFoundError`
 
             ``local_path`` does not found
@@ -188,103 +192,95 @@ class FileUploader(FrozenModel):
         Examples
         --------
 
-        Upload files from ``local_path`` to ``target_path``
+        Upload files from ``local_path`` to ``target_path``:
 
-        .. code:: python
-
-            from onetl.impl import (
-                RemoteFile,
-                LocalPath,
-            )
-            from onetl.file import FileUploader
-
-            uploader = FileUploader(local_path="/local", target_path="/remote", ...)
-            uploaded_files = uploader.run()
-
-            assert uploaded_files.successful == {
+        >>> from onetl.file import FileUploader
+        >>> uploader = FileUploader(local_path="/local", target_path="/remote", ...)
+        >>> upload_result = uploader.run()
+        >>> upload_result
+        UploadResult(
+            successful=FileSet([
                 RemoteFile("/remote/file1"),
                 RemoteFile("/remote/file2"),
-                RemoteFile("/remote/nested/path/file3"),  # directory structure is preserved
-            }
-            assert uploaded_files.failed == {FailedLocalFile("/local/failed.file")}
-            assert uploaded_files.skipped == {LocalPath("/local/already.exists")}
-            assert uploaded_files.missing == {LocalPath("/local/missing.file")}
+                # directory structure is preserved
+                RemoteFile("/remote/nested/path/file3")
+            ]),
+            failed=FileSet([
+                FailedLocalFile("/local/failed.file"),
+            ]),
+            skipped=FileSet([
+                LocalPath("/local/already.exists"),
+            ]),
+            missing=FileSet([
+                LocalPath("/local/missing.file"),
+            ]),
+        )
 
-        Upload only certain files from ``local_path``
+        Upload only certain files from ``local_path``:
 
-        .. code:: python
-
-            from onetl.impl import (
-                RemoteFile,
-                LocalPath,
-            )
-            from onetl.file import FileUploader
-
-            uploader = FileUploader(local_path="/local", target_path="/remote", ...)
-
-            # paths could be relative or absolute, but all should be in "/local"
-            uploaded_files = uploader.run(
-                [
-                    "/local/file1",
-                    "/local/nested/path/file3",
-                    # excluding "/local/file2",
-                ]
-            )
-
-            assert uploaded_files.successful == {
+        >>> from onetl.file import FileUploader
+        >>> uploader = FileUploader(local_path="/local", target_path="/remote", ...)
+        >>> # paths could be relative or absolute, but all should be in "/local"
+        >>> upload_result = uploader.run(
+        ...     [
+        ...         "/local/file1",
+        ...         "/local/nested/path/file3",
+        ...         # excluding "/local/file2",
+        ...     ]
+        ... )
+        >>> upload_result
+        UploadResult(
+            successful=FileSet([
                 RemoteFile("/remote/file1"),
-                RemoteFile("/remote/nested/path/file3"),  # directory structure is preserved
-            }
-            assert not uploaded_files.failed
-            assert not uploaded_files.skipped
-            assert not uploaded_files.missing
+                # directory structure is preserved
+                RemoteFile("/remote/nested/path/file3"),
+            ]),
+            failed=FileSet([]),
+            skipped=FileSet([]),
+            missing=FileSet([]),
+        )
 
-        Upload only certain files from any folder
+        Upload only certain files from any folder:
 
-        .. code:: python
-
-            from onetl.impl import (
-                RemoteFile,
-                LocalPath,
-            )
-            from onetl.file import FileUploader
-
-            uploader = FileUploader(target_path="/remote", ...)  # no local_path set
-
-            # only absolute paths
-            uploaded_files = uploader.run(
-                [
-                    "/local/file1.txt",
-                    "/any/nested/path/file3.txt",
-                ]
-            )
-
-            assert uploaded_files.successful == {
-                RemoteFile("/remote/file1"),
-                RemoteFile("/remote/file3"),
+        >>> from onetl.file import FileUploader
+        >>> uploader = FileUploader(target_path="/remote", ...)  # no local_path set
+        >>> # only absolute paths
+        >>> upload_result = uploader.run(
+        ...     [
+        ...         "/local/file1.txt",
+        ...         "/any/nested/path/file3.txt",
+        ...     ]
+        ... )
+        >>> upload_result
+        UploadResult(
+            successful=FileSet([
+                RemoteFile("/remote/file1.txt"),
                 # directory structure is NOT preserved without local_path
-            }
-            assert not uploaded_files.failed
-            assert not uploaded_files.skipped
-            assert not uploaded_files.missing
+                RemoteFile("/remote/file3.txt"),
+            ]),
+            failed=FileSet([]),
+            skipped=FileSet([]),
+            missing=FileSet([]),
+        )
         """
 
+        entity_boundary_log(log, f"{self.__class__.__name__}.run() starts")
+
         if files is None and not self.local_path:
             raise ValueError("Neither file list nor `local_path` are passed")
 
-        self._log_parameters(files)
+        if not self._connection_checked:
+            self._log_parameters(files)
 
-        # Check everything
-        if self.local_path:
-            self._check_local_path()
-
-        self.connection.check()
-        log_with_indent(log, "")
+            if self.local_path:
+                self._check_local_path()
+            self.connection.check()
+            self._connection_checked = True
 
         self.connection.create_dir(self.target_path)
 
         if files is None:
-            log.info("|%s| File list is not passed to `run` method", self.__class__.__name__)
+            log.debug("|%s| File list is not passed to `run` method", self.__class__.__name__)
             files = self.view_files()
 
         if not files:
@@ -311,6 +307,7 @@ class FileUploader(FrozenModel):
             self._remove_temp_dir(current_temp_dir)
 
         self._log_result(result)
+        entity_boundary_log(log, f"{self.__class__.__name__}.run() ends", char="-")
         return result
 
     @slot
@@ -318,8 +315,10 @@ class FileUploader(FrozenModel):
         """
         Get file list in the ``local_path``. |support_hooks|
 
+        .. versionadded:: 0.3.0
+
         Raises
-        -------
+        ------
         :obj:`onetl.exception.DirectoryNotFoundError`
 
             ``local_path`` does not found
@@ -336,27 +335,26 @@ class FileUploader(FrozenModel):
         Examples
         --------
 
-        View files
+        View files:
 
-        .. code:: python
-
-            from onetl.impl import LocalPath
-            from onetl.file import FileUploader
-
-            uploader = FileUploader(local_path="/local", ...)
-
-            view_files = uploader.view_files()
-
-            assert view_files == {
-                LocalPath("/local/file1.txt"),
-                LocalPath("/local/file3.txt"),
-                LocalPath("/local/nested/path/file3.txt"),
-            }
+        >>> from onetl.file import FileUploader
+        >>> uploader = FileUploader(local_path="/local", ...)
+        >>> uploader.view_files()
+        FileSet([
+            LocalPath("/local/file1.txt"),
+            LocalPath("/local/file3.txt"),
+            LocalPath("/local/nested/path/file3.txt"),
+        ])
         """
 
+        if not self.local_path:
+            raise ValueError("Cannot call `.view_files()` without `local_path`")
+
         log.debug("|Local FS| Getting files list from path '%s'", self.local_path)
 
-        self._check_local_path()
+        if not self._connection_checked:
+            self._check_local_path()
+
         result = FileSet()
 
         try:
@@ -383,8 +381,6 @@ class FileUploader(FrozenModel):
         return RemotePath(temp_path) if temp_path else None
 
     def _log_parameters(self, files: Iterable[str | os.PathLike] | None = None) -> None:
-        entity_boundary_log(log, msg="FileUploader starts")
-
         log.info("|Local FS| -> |%s| Uploading files using parameters:'", self.connection.__class__.__name__)
         log_with_indent(log, "local_path = %s", f"'{self.local_path}'" if self.local_path else "None")
         log_with_indent(log, "target_path = '%s'", self.target_path)
@@ -605,4 +601,3 @@ class FileUploader(FrozenModel):
         log.info("")
         log.info("|%s| Upload result:", self.__class__.__name__)
         log_lines(log, str(result))
-        entity_boundary_log(log, msg=f"{self.__class__.__name__} ends", char="-")
