@@ -1,17 +1,5 @@
-#  Copyright 2023 MTS (Mobile Telesystems)
-#
-#  Licensed under the Apache License, Version 2.0 (the "License");
-#  you may not use this file except in compliance with the License.
-#  You may obtain a copy of the License at
-#
-#      http://www.apache.org/licenses/LICENSE-2.0
-#
-#  Unless required by applicable law or agreed to in writing, software
-#  distributed under the License is distributed on an "AS IS" BASIS,
-#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-#  See the License for the specific language governing permissions and
-#  limitations under the License.
-
+# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
+# SPDX-License-Identifier: Apache-2.0
 from __future__ import annotations
 
 import logging
@@ -26,6 +14,8 @@ def limits_stop_at(path: PathProtocol, limits: Iterable[BaseFileLimit]) -> bool:
     """
     Check if some of limits stops at given path.
 
+    .. versionadded:: 0.8.0
+
     Parameters
     ----------
     path : :obj:`onetl.base.path_protocol.PathProtocol`
@@ -43,15 +33,15 @@ def limits_stop_at(path: PathProtocol, limits: Iterable[BaseFileLimit]) -> bool:
     Examples
     --------
 
-    .. code:: python
-
-        from onetl.file.limit import MaxFilesCount, limits_stop_at
-        from onetl.impl import LocalPath
-
-        limits = [MaxFilesCount(1)]
-
-        assert not limits_stop_at(LocalPath("/path/to/file.csv"), limits)
-        assert limits_stop_at(LocalPath("/path/to/file.csv"), limits)
+    >>> from onetl.file.limit import MaxFilesCount, limits_stop_at
+    >>> from onetl.impl import LocalPath
+    >>> limits = [MaxFilesCount(2)]
+    >>> limits_stop_at(LocalPath("/path/to/file1.csv"), limits)
+    False
+    >>> limits_stop_at(LocalPath("/path/to/file2.csv"), limits)
+    False
+    >>> limits_stop_at(LocalPath("/path/to/file3.csv"), limits)
+    True
     """
     reached = []
     for limit in limits:
