@@ -1,3 +1,5 @@
+# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
+# SPDX-License-Identifier: Apache-2.0
 from __future__ import annotations
 
 import inspect
@@ -8,7 +10,7 @@ from contextlib import ExitStack, suppress
 from functools import partial, wraps
 from typing import Any, Callable, ContextManager, TypeVar
 
-from typing_extensions import ParamSpec, Protocol
+from typing_extensions import Protocol
 
 from onetl.exception import SignatureError
 from onetl.hooks.hook import CanProcessResult, Hook, HookPriority
@@ -17,13 +19,12 @@ from onetl.hooks.hooks_state import HooksState
 from onetl.hooks.method_inheritance_stack import MethodInheritanceStack
 from onetl.log import NOTICE
 
+Method = TypeVar("Method", bound=Callable[..., Any])
+
 logger = logging.getLogger(__name__)
 
-P = ParamSpec("P")
-T = TypeVar("T")
 
-
-def _unwrap_method(method: Callable[P, T]) -> Callable[P, T]:
+def _unwrap_method(method: Method) -> Method:
     """Unwrap @classmethod and @staticmethod to get original function"""
     return getattr(method, "__func__", method)
 
@@ -66,6 +67,8 @@ def bind_hook(method: Callable, inp=None):
 
     See :ref:`hooks-design` for more details.
 
+    .. versionadded:: 0.7.0
+
     Examples
     --------
 
@@ -83,20 +86,20 @@ def bind_hook(method: Callable, inp=None):
 
         @MyClass.method.bind
         @hook
-        def hook(self, arg):
+        def callable(self, arg):
             if arg == "some":
                 do_something()
 
 
         @MyClass.method.bind
         @hook(priority=HookPriority.FIRST, enabled=True)
-        def another_hook(self, arg):
+        def another_callable(self, arg):
             if arg == "another":
                 raise NotAllowed()
 
 
         obj = MyClass()
-        obj.method(1)  # will call both hook(obj, 1) and another_hook(obj, 1)
+        obj.method(1)  # will call both callable(obj, 1) and another_callable(obj, 1)
     """
 
     def inner_wrapper(hook):  # noqa: WPS430
@@ -140,8 +143,7 @@ def _prepare_hook_args(
         @support_hooks
         class MyClass:
             @slot
-            def method(self, some, named="abc"):
-                ...
+            def method(self, some, named="abc"): ...
 
     then hook should have a compatible signature, like these ones:
 
@@ -149,22 +151,19 @@ def _prepare_hook_args(
 
         @MyClass.method.bind
         @hook
-        def callback(self, some, named):
-            ...
+        def callback(self, some, named): ...
 
     .. code:: python
 
         @MyClass.method.bind
         @hook
-        def callback(self, some, **kwargs):
-            ...
+        def callback(self, some, **kwargs): ...
 
     .. code:: python
 
         @MyClass.method.bind
         @hook
-        def callback(my_class_instance, *args, **kwargs):
-            ...
+        def callback(my_class_instance, *args, **kwargs): ...
 
     .. note::
 
@@ -258,6 +257,8 @@ def register_slot(cls: type, method_name: str):  # noqa: WPS231, WPS213, WPS212
     Also ``@classmethod`` is a descriptor, and it can be called only my accessing the class itself,
     which is not possible within a decorator.
 
+    .. versionadded:: 0.7.0
+
     Examples
     --------
 
@@ -468,10 +469,12 @@ def is_slot(method: Callable) -> bool:
 
 
 class Slot(Protocol):
-    """Protocol which is implemented by a method after applying :obj:`~slot` decorator."""
+    """Protocol which is implemented by a method after applying :obj:`~slot` decorator.
 
-    def __call__(self, *args, **kwargs):
-        ...
+    .. versionadded:: 0.7.0
+    """
+
+    def __call__(self, *args, **kwargs): ...
 
     @property
     def __hooks__(self) -> HookCollection:
@@ -562,14 +565,12 @@ class Slot(Protocol):
             @support_hooks
             class MyClass:
                 @slot
-                def my_method(self, arg):
-                    ...
+                def my_method(self, arg): ...
 
 
             @MyClass.my_method.bind
             @hook
-            def callback1(self, arg):
-                ...
+            def callback1(self, arg): ...
 
 
             obj = MyClass()
@@ -599,14 +600,12 @@ class Slot(Protocol):
             @support_hooks
             class MyClass:
                 @slot
-                def my_method(self, arg):
-                    ...
+                def my_method(self, arg): ...
 
 
             @MyClass.my_method.bind
             @hook
-            def callback1(self, arg):
-                ...
+            def callback1(self, arg): ...
 
 
             obj = MyClass()
@@ -620,11 +619,10 @@ class Slot(Protocol):
         """
 
     @wraps(bind_hook)
-    def bind(self):
-        ...
+    def bind(self): ...
 
 
-def slot(method: Callable[P, T]) -> Callable[P, T]:
+def slot(method: Method) -> Method:
     """
     Decorator which enables hooks functionality on a specific class method.
 
@@ -646,8 +644,10 @@ def slot(method: Callable[P, T]) -> Callable[P, T]:
         It is not allowed to use this decorator over ``_private`` and ``__protected`` methods and ``@property``.
         But is allowed to use on ``__dunder__`` methods, like ``__init__``.
 
+    .. versionadded:: 0.7.0
+
     Examples
-    ---------
+    --------
 
     .. code:: python
 
@@ -657,36 +657,30 @@ def slot(method: Callable[P, T]) -> Callable[P, T]:
         @support_hooks
         class MyClass:
             @slot
-            def my_method(self, arg):
-                ...
+            def my_method(self, arg): ...
 
             @slot  # decorator should be on top of all other decorators
             @classmethod
-            def class_method(cls):
-                ...
+            def class_method(cls): ...
 
             @slot  # decorator should be on top of all other decorators
             @staticmethod
-            def static_method(arg):
-                ...
+            def static_method(arg): ...
 
 
         @MyClass.my_method.bind
         @hook
-        def callback1(self, arg):
-            ...
+        def callback1(self, arg): ...
 
 
         @MyClass.class_method.bind
         @hook
-        def callback2(cls):
-            ...
+        def callback2(cls): ...
 
 
         @MyClass.static_method.bind
         @hook
-        def callback3(arg):
-            ...
+        def callback3(arg): ...
 
 
         obj = MyClass()
